from typing import Dict

# Constant is taken from source of PyRMVtransport:
# https://github.com/cgtobi/PyRMVtransport/blob/development/RMVtransport/const.py
PRODUCTS: Dict[str, int] = {
    "ice": 1,
    "ic": 2,
    "ec": 2,
    "r": 4,
    "rb": 4,
    "re": 4,
    "sbahn": 8,
    "ubahn": 16,
    "tram": 32,
    "bus": 64,
    "bus2": 128,
    "ferry": 256,
    "taxi": 512,
    "bahn": 1024,
}


from datetime import datetime, timedelta
from typing import Union
from pyrmv.classes.Board import BoardArrival
from pyrmv.classes.Stop import Stop, StopTrip
from pyrmv.enums.board_type import BoardArrivalType
from pyrmv.raw import board_arrival as raw_board_arrival
from pyrmv.utility.find_exception import find_exception

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

def board_arrival(
    
        access_id: str,
        id: str = None, # type: ignore
        id_ext: str = None, # type: ignore
        direction: Union[str, Stop, StopTrip] = None, # type: ignore
        date: Union[str, datetime] = None, # type: ignore
        time: Union[str, datetime] = None, # type: ignore
        duration: Union[int, timedelta] = 60,
        journeys_max: int = -1,
        operators: Union[str, list] = None, # type: ignore
        lines: Union[str, list] = None, # type: ignore
        passlist: bool = False,
        board_type: Literal[BoardArrivalType.ARR, BoardArrivalType.ARR_EQUIVS, BoardArrivalType.ARR_MAST, BoardArrivalType.ARR_STATION] = BoardArrivalType.ARR,
    ) -> BoardArrival:
    """Method returns a board with arriving transport. 

    More detailed request is available as `raw.board_arrival()`, however returns `dict` instead of `Board`.

    ### Args:
        * access_id (`str`): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * id (`str`, **optional**): Specifies the station/stop ID. Such ID can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * id_ext (`str`, **optional**): Deprecated. Please use `id` as it supports external IDs. Specifies the external station/stop ID. Such ID can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * direction (`Union[str, Stop, StopTrip]`, **optional**): If only vehicles departing or arriving from a certain direction shall be returned, specify the direction by giving the station/stop ID of the last stop on the journey or the Stop object itself. Defaults to `None`.
        * date (`Union[str, datetime]`, **optional**): Sets the start date for which the departures shall be retrieved. Represented in the format YYYY-MM-DD or as a datetime object. By default the current server date is used. Defaults to `None`.
        * time (`Union[str, datetime]`, **optional**): Sets the start time for which the departures shall be retrieved. Represented in the format hh:mm[:ss] in 24h nomenclature or as a datetime object. Seconds will be ignored for requests. By default the current server time is used. Defaults to `None`.
        * duration (`Union[int, timedelta]`, **optional**): Set the interval size in minutes. Can also be provided as a timedelta. Defaults to `60`.
        * journeys_max (`int`, **optional**): Maximum number of journeys to be returned. If no value is defined or -1, all departing/arriving services within the duration sized period are returned. Defaults to `-1`.
        * operators (`Union[str, list]`, **optional**): Only journeys provided by the given operators are part of the result. To filter multiple operators, separate the codes by comma or provide a list. If the operator should not be part of the be trip, negate it by putting ! in front of it. Example: Filter for operator A and B: `operators=[A,B]`. Defaults to `None`.
        * lines (`Union[str, list]`, **optional**): Only journeys running the given line are part of the result. To filter multiple lines, provide a list or separate the codes by comma or provide a list. If the line should not be part of the be trip, negate it by putting ! in front of it. Defaults to `None`.
        * passlist (`bool`, **optional**): Include a list of all passed waystops. Defaults to `False`.
        * board_type (`Union[BoardArrivalType.ARR, BoardArrivalType.ARR_EQUIVS, BoardArrivalType.ARR_MAST, BoardArrivalType.ARR_STATION]`, optional): Set the station arrival board type to be used. Defaults to `BoardArrivalType.ARR`.

    ### Returns:
        * BoardArrival: Instance of BoardArrival object.
    """

    if (isinstance(direction, Stop) or isinstance(direction, StopTrip)):
        direction = direction.id

    board_raw = raw_board_arrival(
        accessId=access_id,
        id=id,
        extId=id_ext,
        direction=direction,
        date=date,
        time=time,
        duration=duration,
        maxJourneys=journeys_max,
        operators=operators,
        lines=lines,
        passlist=passlist,
        boardType=board_type.code
    )

    find_exception(board_raw)

    return BoardArrival(board_raw, access_id)


from requests import get
from typing import OrderedDict, Union
from xmltodict import parse as xmlparse
from datetime import datetime

from pyrmv.utility.weekdays_bitmask import weekdays_bitmask

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# 2.37. HIM Search (himsearch)
def him_search(accessId: str,
        json: bool = True,
        dateB: Union[str, datetime] = None, # type: ignore
        dateE: Union[str, datetime] = None, # type: ignore
        timeB: Union[str, datetime] = None, # type: ignore
        timeE: Union[str, datetime] = None, # type: ignore
        weekdays: Union[str, OrderedDict[str, bool]] = None, # type: ignore
        himIds: Union[str, list] = None, # type: ignore
        hierarchicalView: bool = False,
        operators: Union[str, list] = None, # type: ignore
        categories: Union[str, list] = None, # type: ignore
        channels: Union[str, list] = None, # type: ignore
        companies: Union[str, list] = None, # type: ignore
        lines: Union[str, list] = None, # type: ignore
        lineids: Union[str, list] = None, # type: ignore
        stations: Union[str, list] = None, # type: ignore
        fromstation: str = None, # type: ignore
        tostation: str = None, # type: ignore
        bothways: bool = None, # type: ignore
        trainnames: Union[str, list] = None, # type: ignore
        metas: Union[str, list] = None, # type: ignore
        himcategory: str = None, # type: ignore
        himtags: Union[str, list] = None, # type: ignore
        regions: Union[str, list] = None, # type: ignore
        himtext: Union[str, list] = None, # type: ignore
        himtexttags: Union[str, list] = None, # type: ignore
        additionalfields: Union[str, list, dict] = None, # type: ignore
        poly: bool = False,
        searchmode: Literal["MATCH", "NOMATCH", "TFMATCH"] = None, # type: ignore
        affectedJourneyMode: Literal["ALL", "OFF"] = None, # type: ignore
        affectedJourneyStopMode: Literal["ALL", "IMP", "OFF"] = None, # type: ignore
        orderBy: Union[str, list] = None, # type: ignore
        minprio: Union[str, int] = None, # type: ignore
        maxprio: Union[str, int] = None # type: ignore
    ) -> dict:
    """The himSearch will return a list of HIM messages if matched by the given criteria as well as affected
    products if any.

    Read more about this in section 2.37. "HIM Search (himsearch)" of HAFAS ReST Documentation.  

    ### Args:
        * accessId (str): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * json (bool, optional): Whether response should be retrieved as JSON. XML is returned if False. Defaults to True.
        * dateB (Union[str, datetime], optional): Sets the event period start date. Represented in the format `YYYY-MM-DD`. By default the current server date is used. Defaults to None.
        * dateE (Union[str, datetime], optional): Sets the event period end date. Represented in the format `YYYY-MM-DD`. By default the current server date is used. Defaults to None.
        * timeB (Union[str, datetime], optional): Sets the event period start time. Represented in the format `hh:mm[:ss]` in 24h nomenclature. Seconds will be ignored for requests. By default the current server time is used. Defaults to None.
        * timeE (Union[str, datetime], optional): Sets the event period end time. Represented in the format `hh:mm[:ss]` in 24h nomenclature. Seconds will be ignored for requests. By default the current server time is used. Defaults to None.
        * weekdays (Union[str, OrderedDict[str, bool]], optional): Bitmask or an OrderedDict for validity of HIM messages based on weekdays. OrderedDict must be formatted as follows: `OrderedDict(Monday=bool, Tuesday=bool, Wednesday=bool, Thursday=bool, Friday=bool, Saturday=bool, Sunday=bool)`. Each character of a bitmask represents a weekday starting on monday. Example: Only find HIM Messages valid from monday to friday: `1111100`. Defaults to None.
        * himIds (Union[str, list], optional): List of HIM message IDs as a list or as a string separated by comma. Defaults to None.
        * hierarchicalView (bool, optional): Return parent messages with childs. Defaults to False.
        * operators (Union[str, list], optional): List of operators as a list or as a string separated by comma. Defaults to None.
        * categories (Union[str, list], optional): List of train categories as a list or as a string separated by comma. Defaults to None.
        * channels (Union[str, list], optional): List of channels as a list or as a string separated by comma. Defaults to None.
        * companies (Union[str, list], optional): List of companies as a list or as a string separated by comma. Defaults to None.
        * lines (Union[str, list], optional): Only HIM messages for the given line are part of the result. To filter multiple lines, separate the codes by comma. Defaults to None.
        * lineids (Union[str, list], optional): Only HIM messages for the given line (identified by its line ID) are part of the result. To filter multiple lines, separate the line IDs by comma. Defaults to None.
        * stations (Union[str, list], optional): List of (external) station ids to be filtered for as a list or as a string separated by comma. Defaults to None.
        * fromstation (str, optional): Filter messages by line segment starting at this station given as (external) station id. Defaults to None.
        * tostation (str, optional): Filter messages by line segment travelling in direction of this station given as (external) station id. Defaults to None.
        * bothways (bool, optional): If enabled, messages in both directions - from 'fromstation' to 'tostation' as well as from 'tostation' to 'fromstation' are returned. Defaults to None.
        * trainnames (Union[str, list], optional): List of train name to be filtered for as a list or as a string separated by comma. Defaults to None.
        * metas (Union[str, list], optional): List of predefined filters as a list or as a string separated by comma. Defaults to None.
        * himcategory (str, optional): HIM category, e.g. Works and/or Disturbance. Value depends on your HAFAS server data. Defaults to None.
        * himtags (Union[str, list], optional): HIM Tags. Value depends on your HAFAS server data. Return HIM messages having these tag(s) only. Multiple values are separated by comma. Note: HIM tags differ from HIM text tags. Defaults to None.
        * regions (Union[str, list], optional): Filter for HIM messages based on regions defined in HAFAS raw data. As a list or as a string separated by comma. Available regions can be retrieved by /datainfo service. Defaults to None.
        * himtext (Union[str, list], optional): Filter for HIM messages containing the given free text message as a list or as a string separated by comma. Defaults to None.
        * himtexttags (Union[str, list], optional): Return HIM texts having this text tag(s) only. Multiple values are separated by comma. Note: HIM text tags differ from HIM tags. Defaults to None.
        * additionalfields (Union[str, list, dict], optional): List of additional fields and values to be filtered for. Two filter options are available: Filter by key only: `additionalfields=key` or filter by key and value: `additionalfields={key:value}`. Multiple filters are separated by comma like `additionalfields=[keyA,keyB]` or `additionalfields={key:value}`. Defaults to None.
        * poly (bool, optional): Enables/disables returning of geo information for affected edges and regions if available and enabled in the backend. Defaults to False.
        * searchmode (Literal["MATCH", "NOMATCH", "TFMATCH"], optional): HIM search mode. `"NOMATCH"` iterate over all HIM messages available. `"MATCH"` iterate over all trips to find HIM messages. `"TFMATCH"` uses filters defined `metas` parameter. Defaults to None.
        * affectedJourneyMode (Literal["ALL", "OFF"], optional): Define how to return affected journeys `"OFF"`: do not return affected journeys. `"ALL"`: return affected journeys. Defaults to None.
        * affectedJourneyStopMode (Literal["ALL", "IMP", "OFF"], optional): Define how to return stops of affected journeys. `"IMP"`: return important stops of affected journeys. `"OFF"`: do not return stops of affected journeys. `"ALL"`: return all affected stops of affected journeys. Defaults to None.
        * orderBy (Union[str, list], optional): Define the Order the returned messages by fields and directions. Multiple, string comma separated or list entries are supported. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * minprio (Union[str, int], optional): Filter for HIM messages having at least this priority. Defaults to None.
        * maxprio (Union[str, int], optional): Filter for HIM messages having this priority as maximum. Defaults to None.

    ### Returns:
        * dict: Output from RMV. Dict will contain "errorCode" and "errorText" if exception occurs.
    """    

    if json:
        headers = {"Accept": "application/json"}
    else:
        headers = {"Accept": "application/xml"}

    payload = {}

    for var, val in locals().items():
        if str(var) in ["dateB", "dateE"]:
            if val != None:
                if isinstance(val, datetime):
                    payload[str(var)] = val.strftime("%Y-%m-%d")
                else:
                    payload[str(var)] = val
        elif str(var) in ["timeB", "timeE"]:
            if val != None:
                if isinstance(val, datetime):
                    payload[str(var)] = val.strftime("%H:%M")
                else:
                    payload[str(var)] = val
        elif str(var) == "weekdays":
            if val != None:
                if isinstance(val, OrderedDict):
                    payload[str(var)] = weekdays_bitmask(val)
                else:
                    payload[str(var)] = val
        elif str(var) not in ["json", "headers", "payload"]:
            if val != None:
                payload[str(var)] = val

    output = get("https://www.rmv.de/hapi/himsearch", params=payload, headers=headers)

    # Exceptions checker moved to normal methods
    # and exceptions will no longer raise in raw ones.
    # find_exception(output.json())

    if json:
        return output.json()
    else:
        return xmlparse(output.content)


class SvcParamError(Exception):
    """Exception raised for errors in the input arguments.

    ### Attributes:
        * errorCode: Client error code from HAFAS ReST Request Errors.
        * errorText: Description of an error occurred.
    """

    def __init__(self, errorCode: str, errorText: str):
        self.errorCode = errorCode
        self.errorText = errorText
        super().__init__(self.errorText)

    def __str__(self):
        return f'{self.errorCode} -> {self.errorText}'

class SvcLocationError(Exception):
    """
    Location missing or invalid.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcLocationArrivalError(Exception):
    """
    Arrival location missing or invalid.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcLocationDepartureError(Exception):
    """
    Departure location missing or invalid.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcLocationViaError(Exception):
    """
    Unknown change stop.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcLocationEqualError(Exception):
    """
    Origin/destination or vias are equal.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcLocationNearError(Exception):
    """
    Origin and destination are too close.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcDatetimeError(Exception):
    """
    Date/time are missing or invalid.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcDatetimePeriodError(Exception):
    """
    Date/time are not in timetable or allowed period.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcProductError(Exception):
    """
    Product field missing or invalid.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcContextError(Exception):
    """
    Context is invalid.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcNoResultError(Exception):
    """
    No result found.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcSearchError(Exception):
    """
    Unsuccessful search.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SvcNoMatchError(Exception):
    """
    No match found.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__


from pyrmv.classes.Gis import Gis
from pyrmv.classes.Message import Message
from pyrmv.classes.Stop import StopTrip
from isodate import parse_duration

class Leg():
    """Trip leg object."""    
    
    def __init__(self, data: dict):

        self.origin = StopTrip(data["Origin"])
        self.destination = StopTrip(data["Destination"])

        if "GisRef" in data:
            self.gis = Gis(data["GisRef"]["ref"], data["GisRoute"])
        else:
            self.gis = None

        self.index = data["idx"]
        self.name = data["name"]
        self.type = data["type"]

        if "direction" in data:
            self.direction = data["direction"]
        else:
            self.direction = None

        self.messages = []
        if "Messages" in data:
            for message in data["Messages"]["Message"]:
                self.messages.append(Message(message))

        if "number" in data:
            self.number = data["number"]
        else:
            self.number = None

        self.duration = parse_duration(data["duration"])
        
        if "dist" in data:
            self.distance = data["dist"]
        else:
            self.distance = None


class IntError(Exception):
    """
    Internal RMV server error.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class IntGatewayError(Exception):
    """
    Communication error with RMV backend systems.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class IntTimeoutError(Exception):
    """
    Timeout during service processing.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__


from pyrmv.classes.Stop import Stop
from pyrmv.classes.Message import Message

class Journey():
    """Journey object."""    

    def __init__(self, data: dict):
        self.stops = []
        self.direction = data["Directions"]["Direction"][0]["value"]
        self.direction_flag = data["Directions"]["Direction"][0]["flag"]
        self.messages = []

        for stop in data["Stops"]["Stop"]:
            self.stops.append(Stop(stop))

        for message in data["Messages"]["Message"]:
            self.messages.append(Message(message))

    def __str__(self) -> str:
        return f"Journey with total of {len(self.stops)} stops and {len(self.messages)} messages heading {self.direction} ({self.direction_flag})"


from typing import List, Union
from pyrmv.classes.Stop import Stop
from pyrmv.enums.lang import Language
from pyrmv.raw.stop_by_name import stop_by_name as raw_stop_by_name
from pyrmv.utility.find_exception import find_exception

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

def stop_by_id(
    
        access_id: str,
        query: str,
        lang: Literal[Language.DE, Language.DA, Language.EN, Language.ES, Language.FR, Language.HU, Language.IT, Language.NL, Language.NO, Language.PL, Language.SV, Language.TR] = Language.EN,
    ) -> Union[Stop, None]:
    """Method can be used to get Stop object whilst only having id available.

    ### Args:
        * access_id (`str`): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * query (`str`): Search for that token.
        * lang (`Literal[Language.DE, Language.DA, Language.EN, Language.ES, Language.FR, Language.HU, Language.IT, Language.NL, Language.NO, Language.PL, Language.SV, Language.TR]`, **optional**): The language of response. Defaults to `Language.EN`.
    
    ### Returns:
        * Stop: Instance of Stop object or None if not found.
    """    

    stops_raw = raw_stop_by_name(
        accessId=access_id,
        inputString=query,
        lang=lang.code,
        maxNo=1
    )

    find_exception(stops_raw)

    if len(stops_raw["stopLocationOrCoordLocation"]) > 0:
        stop = stops_raw["stopLocationOrCoordLocation"][0]

        if "StopLocation" in stop:
            return Stop(stop["StopLocation"])
        elif "CoordLocation" in stop:
            return Stop(stop["CoordLocation"])
        else:
            return None
    else:
        return 0


class ApiAuthError(Exception):
    """
    Access denied for accessId provided.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class ApiQuotaError(Exception):
    """
    Quota exceeded for accessId provided.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class ApiTooManyRequests(Exception):
    """
    Too many requests.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class ApiParamError(Exception):
    """Exception raised for errors in the input arguments.

    ### Attributes:
        * errorCode: Client error code from HAFAS ReST Request Errors.
        * errorText: Description of an error occurred.
    """

    def __init__(self, errorCode: str, errorText: str):
        self.errorCode = errorCode
        self.errorText = errorText
        super().__init__(self.errorText)

    def __str__(self):
        return f'{self.errorCode} -> {self.errorText}'

class ApiFormatError(Exception):
    """
    Response format not supported.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__


from requests import get
from typing import Union
from xmltodict import parse as xmlparse

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# 2.3. Location Search by Name (location.name)
def stop_by_name(accessId: str,
        inputString: str,
        lang: Literal["de", "da", "en", "es", "fr", "hu", "it", "nl", "no", "pl", "sv", "tr"] = "en",
        json: bool = True,
        maxNo: int = 10,
        stopType: Literal["A", "ALL", "AP", "P", "S", "SA", "SP"] = "ALL",
        locationSelectionMode: Literal["SLCT_N", "SLCT_A"] = None, # type: ignore
        products: int = None, # type: ignore
        coordLat: Union[str, float] = None, # type: ignore
        coordLong: Union[str, float] = None, # type: ignore
        radius: Union[int, float] = 1000,
        refineId: str = None, # type: ignore
        meta: str = None, # type: ignore
        stations: Union[str, list] = None, # type: ignore
        sattributes: Union[str, list] = None, # type: ignore
        filterMode: Literal["DIST_PERI", "EXCL_PERI", "SLCT_PERI"] = "DIST_PERI"
    ) -> dict:
    """The location.name service can be used to perform a pattern matching of a user input and to retrieve a list
    of possible matches in the journey planner database. Possible matches might be stops/stations, points of
    interest and addresses.  

    The result is a list of possible matches (locations) where the user might pick one entry to perform a trip
    request with this location as origin or destination or to ask for a departure board or arrival board of this
    location (stops/stations only).  

    Read more about this in section 2.3. "Location Search by Name (location.name)" of HAFAS ReST Documentation.  

    ### Args:
        * accessId (str): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * inputString (str): Search for that token.
        * lang (Literal["de","da","en","es","fr","hu","it","nl","no","pl","sv","tr"], optional): The language of response. Defaults to "en".
        * json (bool, optional): Whether response should be retrieved as JSON. XML is returned if False. Only matters if raw_response is True. Defaults to True.
        * maxNo (int, optional): Maximum number of returned stops. In range 1-1000. Defaults to 10.
        * stopType (Literal["A", "ALL", "AP", "P", "S", "SA", "SP"], optional): Type filter for location types. Defaults to "ALL".
        * locationSelectionMode (str, optional): Selection mode for locations. "SLCT_N": Not selectable, "SLCT_A": Selectable. Defaults to None.
        * products (int, optional): Decimal value defining the product classes to be included in the search. It represents a bitmask combining bit number of a product as defined in the HAFAS raw data. Defaults to None.
        * coordLat (Union[str, float], optional): Latitude of centre coordinate. Defaults to None.
        * coordLong (Union[str, float], optional): Longitude of centre coordinate. Defaults to None.
        * radius (Union[int, float], optional): Search radius in meter around the given coordinate if any. Defaults to 1000.
        * refineId (str, optional): In case of an refinable location, this value takes the ID of the refinable one of a previous result. Defaults to None.
        * meta (str, optional): Filter by a predefined meta filter. If the rules of the predefined filter should not be negated, put ! in front of it. Defaults to None.
        * stations (Union[str, list], optional): Filter for stations. Matches if the given value is prefix of any station in the itinerary. Multiple values are separated by comma. Defaults to None.
        * sattributes (Union[str, list], optional): Filter locations by one or more attribute codes. Multiple attribute codes are separated by comma. If the attribute should not be part of the be location data, negate it by putting ! in front of it. Defaults to None.
        * filterMode (Literal["DIST_PERI", "EXCL_PERI", "SLCT_PERI"], optional): Filter modes for nearby searches. Defaults to "DIST_PERI".

    ### Returns:
        * dict: Output from RMV. Dict will contain "errorCode" and "errorText" if exception occurs.
    """    

    if json:
        headers = {"Accept": "application/json"}
    else:
        headers = {"Accept": "application/xml"}

    payload = {}

    for var, val in locals().items():
        if str(var) == "inputString":
            if val != None:
                payload["input"] = val
        elif str(var) == "stopType":
            if val != None:
                payload["type"] = val.upper()
        elif str(var) == "filterMode":
            if val != None:
                payload["filterMode"] = val.upper()
        elif str(var) not in ["json", "headers", "payload", "raw_response", "stopType"]:
            if val != None:
                payload[str(var)] = val

    output = get("https://www.rmv.de/hapi/location.name", params=payload, headers=headers)

    # Exceptions checker moved to normal methods
    # and exceptions will no longer raise in raw ones.
    # find_exception(output.json())

    if json:
        return output.json()
    else:
        return xmlparse(output.content)


from datetime import datetime


class Stop():
    """Stop object."""    
    
    def __init__(self, data: dict):

        self.name = data["name"]
        self.id = data["id"]

        if "extId" in data:
            self.ext_id = data["extId"]
        else:
            self.ext_id = None

        if "description" in data:
            self.description = data["description"]
        else:
            self.description = None

        self.lon = data["lon"]
        self.lat = data["lat"]

        if "routeIdx" in data:
            self.route_index = data["routeIdx"]
        else:
            self.route_index = None

        if "arrTrack" in data:
            self.track_arrival = data["arrTrack"]
        else:
            self.track_arrival = None

        if "depTrack" in data:
            self.track_departure = data["depTrack"]
        else:
            self.track_departure = None

    def __str__(self) -> str:
        return f"Stop {self.name} at {self.lon}, {self.lat}"

class StopTrip(Stop):
    """Trip stop object. It's like a Stop object, but with a date and time."""    

    def __init__(self, data: dict):
        
        self.type = data["type"]
        self.date = datetime.strptime(data["date"], "%Y-%m-%d")
        self.time = datetime.strptime(data["time"], "%H:%M:%S")
        super().__init__(data)

    def __str__(self) -> str:
        return f"Stop {self.name} at {self.lon}, {self.lat} at {self.time.time()} {self.date.date()}"


from datetime import datetime
from requests import get
from typing import List, Union
from xmltodict import parse as xmlparse

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# 2.12. Trip Search (trip) 
def trip_find(accessId: str,
        lang: Literal["de", "da", "en", "es", "fr", "hu", "it", "nl", "no", "pl", "sv", "tr"] = "en",
        json: bool = True,

        originId: str = None, # type: ignore
        originExtId: str = None, # type: ignore
        originCoordLat: Union[str, float] = None, # type: ignore
        originCoordLong: Union[str, float] = None, # type: ignore
        originCoordName: str = None, # type: ignore

        destId: str = None, # type: ignore
        destExtId: str = None, # type: ignore
        destCoordLat: Union[str, float] = None, # type: ignore
        destCoordLong: Union[str, float] = None, # type: ignore
        destCoordName: str = None, # type: ignore

        via: str = None, # type: ignore
        viaId: str = None, # type: ignore
        viaWaitTime: int = 0,

        avoid: str = None, # type: ignore
        avoidId: str = None, # type: ignore

        viaGis: str = None, # type: ignore

        changeTimePercent: int = 100,
        minChangeTime: int = None, # type: ignore
        maxChangeTime: int = None, # type: ignore
        addChangeTime: int = None, # type: ignore
        maxChange: int = None, # type: ignore

        date: Union[str, datetime] = None, # type: ignore
        time: Union[str, datetime] = None, # type: ignore
        
        searchForArrival: bool = False,

        numF: int = None, # type: ignore
        numB: int = None, # type: ignore

        context: str = None, # type: ignore

        poly: bool = False,
        polyEnc: Literal["DLT", "GPA", "N"] = "N",

        passlist: bool = False,
        products: int = None, # type: ignore
        operators: Union[str, list] = None, # type: ignore
        
        attributes: Union[str, list] = None, # type: ignore
        sattributes: Union[str, list] = None, # type: ignore
        fattributes: Union[str, list] = None, # type: ignore
        lines: Union[str, list] = None, # type: ignore
        lineids: Union[str, list] = None, # type: ignore

        avoidPaths: List[Literal["SW", "EA", "ES", "RA", "CB"]] = None, # type: ignore

        originWalk: Union[str, list] = None, # type: ignore
        originBike: Union[str, list] = None, # type: ignore
        originCar: Union[str, list] = None, # type: ignore
        originTaxi: Union[str, list] = None, # type: ignore
        originPark: Union[str, list] = None, # type: ignore
        originMeta: Union[str, list] = None, # type: ignore

        destWalk: Union[str, list] = None, # type: ignore
        destBike: Union[str, list] = None, # type: ignore
        destCar: Union[str, list] = None, # type: ignore
        destTaxi: Union[str, list] = None, # type: ignore
        destPark: Union[str, list] = None, # type: ignore
        destMeta: Union[str, list] = None, # type: ignore

        totalWalk: Union[str, list] = None, # type: ignore
        totalBike: Union[str, list] = None, # type: ignore
        totalCar: Union[str, list] = None, # type: ignore
        totalTaxi: Union[str, list] = None, # type: ignore
        totalMeta: Union[str, list] = None, # type: ignore

        gisProducts: str = None, # type: ignore

        includeIv: bool = False,
        ivOnly: bool = False,

        mobilityProfile: str = None, # type: ignore
        
        bikeCarriage: bool = False,
        bikeCarriageType: Literal["SINGLEBIKES", "SMALLGROUPS", "LARGEGROUPS"] = None, # type: ignore

        sleepingCar: bool = False,
        couchetteCoach: bool = False,
        showPassingPoints: bool = False,
        baim: bool = False,

        eco: bool = False,
        ecoCmp: bool = False,
        ecoParams: str = None, # type: ignore

        rtMode: Literal["FULL", "INFOS", "OFF", "REALTIME", "SERVER_DEFAULT"] = None, # type: ignore

        unsharp: bool = False,
        trainFilter: str = None, # type: ignore
        economic: bool = False,
        groupFilter: str = None, # type: ignore

        blockingList: str = None, # type: ignore
        blockedEdges: str = None, # type: ignore

        trainComposition: bool = False,
        includeEarlier: bool = False,
        withICTAlternatives: bool = False,
        tariff: bool = None, # type: ignore
        trafficMessages: bool = False,
        travellerProfileData: str = None, # type: ignore
        withFreq: bool = True
    ) -> dict:
    """The trip service calculates a trip from a specified origin to a specified destination. These might be
    stop/station IDs or coordinates based on addresses and points of interest validated by the location service or
    coordinates freely defined by the client.  

    Read more about this in section 2.12. "Trip Search (trip)" of HAFAS ReST Documentation.  

    ### Args:
        * accessId (str): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * lang (Literal["de","da","en","es","fr","hu","it","nl","no","pl","sv","tr"], optional): The language of response. Defaults to "en".
        * json (bool, optional): Whether response should be retrieved as JSON. XML is returned if False. Defaults to True.
        * originId (str, optional): Specifies the station/stop ID of the origin for the trip. Such ID can be retrieved from stopByName() or stopByCoords(). Defaults to None.
        * originExtId (str, optional): Deprecated. Please use originId as it supports external IDs. Specifies the external station/stop ID of the origin for the trip. Such ID can be retrieved from stopByName() or stopByCoords(). Defaults to None.
        * originCoordLat (Union[str, float], optional): Latitude of station/stop coordinate of the trip's origin. The coordinate can be retrieved from stopByName() or stopByCoords(). Defaults to None.
        * originCoordLong (Union[str, float], optional): Longitude of station/stop coordinate of the trip's origin. The coordinate can be retrieved from stopByName() or stopByCoords(). Defaults to None.
        * originCoordName (str, optional): Name of the trip's origin if coordinate cannot be resolved to an address or poi. Defaults to None.
        * destId (str, optional): Specifies the station/stop ID of the destination for the trip. Such ID can be retrieved from stopByName() or stopByCoords(). Defaults to None.
        * destExtId (str, optional): Deprecated. Please use destId as it supports external IDs. Specifies the external station/stop ID of the destination for the trip. Such ID can be retrieved from stopByName() or stopByCoords(). Defaults to None.
        * destCoordLat (Union[str, float], optional): Latitude of station/stop coordinate of the trip's destination. The coordinate can be retrieved from stopByName() or stopByCoords(). Defaults to None.
        * destCoordLong (Union[str, float], optional): Longitude of station/stop coordinate of the trip's destination. The coordinate can be retrieved from stopByName() or stopByCoords(). Defaults to None.
        * destCoordName (str, optional): Name of the trip's destination if coordinate cannot be resolved to an address or poi. Defaults to None.
        * via (str, optional): Complex structure to provide multiple via points separated by semicolon. This structure is build like this: `viaId|waittime|viastatus|products|direct|sleepingCar|couchetteCoach|attributes`. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * viaId (str, optional): ID of a station/stop used as a via for the trip. Specifying a via station forces the trip search to look for trips which must pass through this station. Such ID can be retrieved from stopByName() or stopByCoords(). If `via` is used, `viaId` and `viaWaitTime ` are having no effect. Defaults to None.
        * viaWaitTime (int, optional): Defines the waiting time spent at via station in minutes. If `via` is used, `viaId` and `viaWaitTime` are having no effect. Defaults to 0.
        * avoid (str, optional): Complex structure to provide multiple points to be avoided separated by semicolon. This structure is build like this: `avoidId|avoidstatus` avoidId: id, extId or altId of the avoid, mandatory avoidstatus: one of NPAVM (do not run through if this is a meta station), NPAVO (do not run through), NCAVM (do not change if this is a meta station), NCAVO (do not change), optional but defaults to NCAVM Example: Just define three avoids by extId: `avoid="801234;801235;801236"`. Defaults to None.
        * avoidId (str, optional): ID of a station/stop to be avoided as transfer stop for the trip. Such ID can be retrieved from stopByName() or stopByCoords(). If `avoid` is used, `avoidId` has no effect. Defaults to None.
        * viaGis (str, optional): Complex structure to provide multiple GIS via locations separated by semicolon. This structure is build like this: `locationId|locationMode|transportMode|placeType|usageType|mode|durationOfStay`. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * changeTimePercent (int, optional): Configures the walking speed when changing from one leg of the journey to the next one. It extends the time required for changes by a specified percentage. A value of 200 doubles the change time as initially calculated by the system. In the response, change time is presented in full minutes. If the calculation based on changeTime-Percent does not result in a full minute, it is rounded using "round half up" method. Defaults to 100.
        * minChangeTime (int, optional): Minimum change time at stop in minutes. Defaults to None.
        * maxChangeTime (int, optional): Maximum change time at stop in minutes. Defaults to None.
        * addChangeTime (int, optional): This amount of minutes is added to the change time at each stop. Defaults to None.
        * maxChange (int, optional): Maximum number of changes. In range 0-11. Defaults to None.
        * date (Union[str, datetime], optional): Sets the start date for which the departures shall be retrieved. Represented in the format `YYYY-MM-DD`. By default the current server date is used. Defaults to None.
        * time (Union[str, datetime], optional): Sets the start time for which the departures shall be retrieved. Represented in the format `hh:mm[:ss]` in 24h nomenclature. Seconds will be ignored for requests. By default the current server time is used. Defaults to None.
        * searchForArrival (bool, optional): If set, the date and time parameters specify the arrival time for the trip search instead of the departure time. Defaults to False.
        * numF (int, optional): Minimum number of trips after the search time. Sum of numF and numB has to be less or equal 6. Read more about this in HAFAS ReST Documentation. In range 1-6. Defaults to None.
        * numB (int, optional): Minimum number of trips before the search time. Sum of numF and numB has to be less or equal 6. Read more about this in HAFAS ReST Documentation. In range 0-6. Defaults to None.
        * context (str, optional): Defines the starting point for the scroll back or forth operation. Use the scrB value from a previous result to scroll backwards in time and use the scrF value to scroll forth. Defaults to None.
        * poly (bool, optional): Enables/disables the calculation of the polyline for each leg of the trip except any GIS route. Defaults to False.
        * polyEnc (Literal["DLT", "GPA", "N"], optional): Defines encoding of the returned polyline. Possible values are "N" (no encoding / compression), "DLT" (delta to the previous coordinate), "GPA" (Google encoded polyline format). Defaults to "N".
        * passlist (bool, optional): Enables/disables the return of the passlist for each leg of the trip. Defaults to False.
        * products (str, optional): Decimal value defining the product classes to be included in the search. It represents a bitmask combining bit number of a product as defined in the HAFAS raw data. Defaults to None.
        * operators (Union[str, list], optional): Only trips provided by the given operators are part of the result. If the operator should not be part of the be trip, negate it by putting ! in front of it. Example: Filter for operator A and B: `operators=["A","B"]`. Defaults to None.
        * attributes (Union[str, list], optional): Filter trips by one or more attribute codes of a journey. If the attribute should not be part of the be trip, negate it by putting ! in front of it. Defaults to None.
        * sattributes (Union[str, list], optional): Filter trips by one or more station attribute codes of a journey. If the attribute should not be part of the be trip, negate it by putting ! in front of it. Defaults to None.
        * fattributes (Union[str, list], optional): Filter trips by one or more footway attribute codes of a journey. If the attribute should not be part of the be trip, negate it by putting ! in front of it. Defaults to None.
        * lines (Union[str, list], optional): Only journeys running the given line are part of the result. If the line should not be part of the be trip, negate it by putting ! in front of it. Defaults to None.
        * lineids (Union[str, list], optional): Only journeys running the given line (identified by its line ID) are part of the result. If the line should not be part of the be trip, negate it by putting ! in front of it. Defaults to None.
        * avoidPaths (List[Literal["SW", "EA", "ES", "RA", "CB"]], optional): Only path not having the given properties will be part of the result. "SW": Stairway; "EA": Elevator; "ES": Escalator; "RA": Ramp; "CB": Convey Belt. Example: Use paths without ramp and stairway: `avoidPaths="SW", "RA"`. Defaults to None.
        * originWalk (Union[str, list], optional): Enables/disables using footpaths in the beginning of a trip when searching from an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * originBike (Union[str, list], optional): Enables/disables using bike routes in the beginning of a trip when searching from an address. To fine-tune the minimum and/or maximum distance to the next public transport station or mode change point, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * originCar (Union[str, list], optional): Enables/disables using car in the beginning of a trip when searching from an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * originTaxi (Union[str, list], optional): Enables/disables using taxi rides in the beginning of a trip when searching from an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * originPark (Union[str, list], optional): Enables/disables using Park and Ride in the beginning of a trip when searching from an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * originMeta (Union[str, list], optional): Enables using one or more predefined individual transport meta profile at the beginning of a trip. Defaults to None.
        * destWalk (Union[str, list], optional): Enables/disables using footpaths at the end of a trip when searching to an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * destBike (Union[str, list], optional): Enables/disables using bike routes at the end of a trip when searching to an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * destCar (Union[str, list], optional): Enables/disables using car routes at the end of a trip when searching to an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * destTaxi (Union[str, list], optional): Enables/disables using taxi rides at the end of a trip when searching to an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * destPark (Union[str, list], optional): Enables/disables using Park and Ride at the end of a trip when searching to an address. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * destMeta (Union[str, list], optional): Enables using one or more predefined individual transport meta profile at the end of a trip. Defaults to None.
        * totalWalk (Union[str, list], optional): Enables/disables using footpaths for the whole trip. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * totalBike (Union[str, list], optional): Enables/disables using bike routes for the whole trip. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * totalCar (Union[str, list], optional): Enables/disables using car routes for the whole trip. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * totalTaxi (Union[str, list], optional): Enables/disables using taxi rides for the whole trip. To fine-tune the minimum and/or maximum distance to the next public transport station, provide these values as a list or as a string separated by comma. These values are expressed in meters. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * totalMeta (Union[str, list], optional): Enables using one or more predefined individual transport meta profile for a trip. Defaults to None.
        * gisProducts (str, optional): Filter on GIS product, e.g. specific sharing provider. Currently, only exclusion of certain providers is available by adding ! in front of the provider meta code. Defaults to None.
        * includeIv (bool, optional): Enables/disables search for individual transport routes. Defaults to False.
        * ivOnly (bool, optional): Enables/disables search for individual transport routes only. Defaults to False.
        * mobilityProfile (str, optional): Use a predefined filter by its name. The filters are defined in the HAFAS installation. If the filter should be negated, put a ! in front of its name. Defaults to None.
        * bikeCarriage (bool, optional): Enables/disables search for trips explicit allowing bike carriage. This will only work in combination with `maxChange=0` as those trips are always meant to be direct connections. Defaults to False.
        * bikeCarriageType (Literal["SINGLEBIKES", "SMALLGROUPS", "LARGEGROUPS"], optional): Filter for a specific bike carriage type. Defaults to None.
        * sleepingCar (bool, optional): Enables/disables search for trips having sleeping car. This will only work in combination with `maxChange=0` as those trips are always meant to be direct connections. Defaults to False.
        * couchetteCoach (bool, optional): Enables/disables search for trips having couchette coach. This will only work in combination with `maxChange=0` as those trips are always meant to be direct connections. Defaults to False.
        * showPassingPoints (bool, optional): Enables/disables the return of stops having no alighting and boarding in its passlist for each leg of the trip. Needs passlist enabled. Defaults to False.
        * baim (bool, optional): Enables/disables BAIM search and response. Defaults to False.
        * eco (bool, optional): Enables/disables eco value calculation. Defaults to False.
        * ecoCmp (bool, optional): Enables/disables eco comparison. Defaults to False.
        * ecoParams (str, optional): Provide additional eco parameters. Defaults to None.
        * rtMode (Literal["FULL", "INFOS", "OFF", "REALTIME", "SERVER_DEFAULT"], optional): Set the realtime mode to be used. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * unsharp (bool, optional): Enables/disables unsharp search mode. Read more about this in section 2.12.2.1. "Trip Search (trip)" of HAFAS ReST Documentation. Defaults to False.
        * trainFilter (str, optional): Filters a trip search for a certain train. First hit will be taken. Defaults to None.
        * economic (bool, optional): Enables/disables economic search mode. Read more about this in section 2.12.2.2. "Trip Search (trip)" of HAFAS ReST Documentation. Defaults to False.
        * groupFilter (str, optional): Use a predefined group filter to query for certain modes. Defaults to None.
        * blockingList (str, optional): Defines a section of a route of a journey not to be used within the trip search. Each route section is defined by a tuple of the following style: `<train name>|<departure id>|<arrival id>|<departure time>|<arrival time>|<departure date>|<arrival date>` A set of tuples can be separated by semicolon. Defaults to None.
        * blockedEdges (str, optional): List of edges within the public transport network that should be excluded from the result. Each edge is defined by a tuple of the following style: `start location ID|end locationID|bidirectional|blockOnlyIfInOutAllowed` A set of tuples can be separated by semicolon. Defaults to None.
        * trainComposition (bool, optional): Enables/disables train composition data. Defaults to False.
        * includeEarlier (bool, optional): Disables search optimization in relation of duration. Defaults to False.
        * withICTAlternatives (bool, optional): Enables/disables the search for alternatives with individualized change times (ICT). Defaults to False.
        * tariff (bool, optional): Enables/disables the output of tariff data. The default is configurable via provisioning. Defaults to None.
        * trafficMessages (bool, optional): Enables/disables the output of traffic messages. The default is configurable via provisioning. Defaults to False.
        * travellerProfileData (str, optional): Traveller profile data. Structure depends on set up. Defaults to None.
        * withFreq (bool, optional): Enables/disables the calculation of frequency information. Defaults to True.

    ### Returns:
        * dict: Output from RMV. Dict will contain "errorCode" and "errorText" if exception occurs.
    """    

    if json:
        headers = {"Accept": "application/json"}
    else:
        headers = {"Accept": "application/xml"}

    payload = {}

    for var, val in locals().items():
        if str(var) == "date":
            if val != None:
                if isinstance(val, datetime):
                    payload[str(var)] = val.strftime("%Y-%m-%d")
                else:
                    payload[str(var)] = val
        elif str(var) == "time":
            if val != None:
                if isinstance(val, datetime):
                    payload[str(var)] = val.strftime("%H:%M")
                else:
                    payload[str(var)] = val
        elif str(var) == "rtMode":
            if val != None:
                payload["rtMode"] = val.upper()
        elif str(var) not in ["json", "headers", "payload"]:
            if val != None:
                payload[str(var)] = val

    output = get("https://www.rmv.de/hapi/trip", params=payload, headers=headers)

    # Exceptions checker moved to normal methods
    # and exceptions will no longer raise in raw ones.
    # find_exception(output.json())

    if json:
        return output.json()
    else:
        return xmlparse(output.content)


"""
## PythonRMV

Small module that makes your journey with RMV REST API somehow easier. Based fully on official RMV API reference and HAFAS documentation.

## Usage

```py
import pyrmv

# Set API key
access_id = "Something"

# Get origin's and destination's location
origin = pyrmv.stop_by_name(access_id, "Frankfurt Hauptbahnhof", max_number=3)[0]
destination = pyrmv.stop_by_coords(access_id, 50.099613, 8.685449, max_number=3)[0]

# Find a trip by locations got
trip = pyrmv.trip_find(access_id, origin_id=origin.id, dest_id=destination.id)
```
"""

__name__ = "pyrmv"
__version__ = "0.2.9"
__license__ = "MIT License"
__author__ = "Profitroll"

from . import raw
from . import const
from . import enums
from . import errors
from . import utility
from . import classes
from .methods import *


from enum import auto
from .auto_name import AutoName

class Language(AutoName):
    """Enumeration used to declare locales as ISO-3166 codes (but only available in HAFAS ones)."""

    DE = auto()
    "German"

    DA = auto()
    "Danish"

    EN = auto()
    "English"

    ES = auto()
    "Spanish"

    FR = auto()
    "French"

    HU = auto()
    "Hungarian"

    IT = auto()
    "Italian"

    NL = auto()
    "Dutch"

    NO = auto()
    "Norwegian"

    PL = auto()
    "Polish"

    SV = auto()
    "Swedish"

    TR = auto()
    "Turkish"


from enum import auto
from .auto_name import AutoName

class SelectionMode(AutoName):
    """Enumeration used to declare location selection modes.
    
    * SLCT_A - Selectable
    * SLCT_N - Not selectable
    """

    SLCT_A = auto()
    "Selectable"

    SLCT_N = auto()
    "Not selectable"


from requests import get
from typing import Union
from xmltodict import parse as xmlparse

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# 2.4. Location Search by Coordinate (location.nearbystops)
def stop_by_coords(accessId: str,
        originCoordLat: Union[str, float],
        originCoordLong: Union[str, float],
        lang: Literal["de", "da", "en", "es", "fr", "hu", "it", "nl", "no", "pl", "sv", "tr"] = "en",
        json: bool = True,
        radius: Union[int, float] = 1000,
        maxNo: int = 10,
        stopType: Literal["S", "P", "SP", "SE", "SPE"] = "S",
        locationSelectionMode: Literal["SLCT_N", "SLCT_A"] = None, # type: ignore
        products: int = None, # type: ignore
        meta: str = None, # type: ignore
        sattributes: Union[str, list] = None, # type: ignore
        sinfotexts: Union[str, list] = None # type: ignore
    ) -> dict:
    """The location.nearbystops service returns a list of stops around a given center coordinate (within a
    radius of 1000m). The returned results are ordered by their distance to the center coordinate.  

    Read more about this in section 2.4. "Location Search by Coordinate (location.nearbystops)" of HAFAS ReST Documentation.  

    ### Args:
        * accessId (str): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * originCoordLat (Union[str, float]): Latitude of centre coordinate.
        * originCoordLong (Union[str, float]): Longitude of centre coordinate.
        * lang (Literal["de","da","en","es","fr","hu","it","nl","no","pl","sv","tr"], optional): The language of response. Defaults to "en".
        * json (bool, optional): Whether response should be retrieved as JSON. XML is returned if False. Only matters if raw_response is True. Defaults to True.
        * radius (Union[int, float], optional): Search radius in meter around the given coordinate if any. Defaults to 1000.
        * maxNo (int, optional): Maximum number of returned stops. Defaults to 10.
        * stopType (Literal["S", "P", "SP", "SE", "SPE"], optional): Type filter for location types. Defaults to "S".
        * locationSelectionMode (Literal["SLCT_N", "SLCT_A"], optional): Selection mode for locations. Defaults to None.
        * products (int, optional): Decimal value defining the product classes to be included in the search. It represents a bitmask combining bit number of a product as defined in the HAFAS raw data. Defaults to None.
        * meta (str, optional): Filter by a predefined meta filter. If the rules of the predefined filter should not be negated, put ! in front of it. Defaults to None.
        * sattributes (Union[str, list], optional): Filter locations by one or more attribute codes. Multiple attribute codes are separated by comma. If the attribute should not be part of the be location data, negate it by putting ! in front of it. Defaults to None.
        * sinfotexts (Union[str, list], optional): Filter locations by one or more station infotext codes and values. Multiple attribute codes are separated by comma the value by pipe |. Defaults to None.

    ### Returns:
        * dict: Output from RMV. Dict will contain "errorCode" and "errorText" if exception occurs.
    """    

    if json:
        headers = {"Accept": "application/json"}
    else:
        headers = {"Accept": "application/xml"}

    payload = {}

    for var, val in locals().items():
        if str(var) == "stopType":
            if val != None:
                payload["type"] = val.upper()
        elif str(var) == "locationSelectionMode":
            if val != None:
                payload["locationSelectionMode"] = val.upper()
        elif str(var) not in ["json", "headers", "payload", "raw_response"]:
            if val != None:
                payload[str(var)] = val

    output = get("https://www.rmv.de/hapi/location.nearbystops", params=payload, headers=headers)

    # Exceptions checker moved to normal methods
    # and exceptions will no longer raise in raw ones.
    # find_exception(output.json())

    if json:
        return output.json()
    else:
        return xmlparse(output.content)


from requests import get
from datetime import datetime, timedelta
from typing import Union
from xmltodict import parse as xmlparse

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# 2.24. Departure Board (departureBoard)
def board_departure(accessId: str,
        json: bool = True,
        id: str = None, # type: ignore
        extId: str = None, # type: ignore
        direction: str = None, # type: ignore
        date: Union[str, datetime] = None, # type: ignore
        time: Union[str, datetime] = None, # type: ignore
        duration: int = 60,
        maxJourneys: int = -1,
        products: int = None, # type: ignore
        operators: Union[str, list] = None, # type: ignore
        lines: Union[str, list] = None, # type: ignore
        filterEquiv: bool = True,
        attributes: Union[str, list] = None, # type: ignore
        platforms: Union[str, list] = None, # type: ignore
        passlist: bool = False,
        boardType: Literal["DEP", "DEP_EQUIVS", "DEP_MAST", "DEP_STATION"] = "DEP"
    ) -> dict:
    """The separture board can be retrieved by a call to the departureBoard service. This method will return the
    next departures (or less if not existing) from a given point in time within a duration covered time span. The
    default duration size is 60 minutes.  

    Note: The result list always contains all departures running the the last minute found even if the requested
    maximum was overrun.  

    Read more about this in section 2.24. "Departure Board (departureBoard)" of HAFAS ReST Documentation.  

    ### Args:
        * accessId (str): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * json (bool, optional): Whether response should be retrieved as JSON. XML is returned if False. Only matters if raw_response is True. Defaults to True.
        * id (str, optional): Access ID for identifying the requesting client. Defaults to None.
        * extId (str, optional): Deprecated. Please use id as it supports external IDs. Specifies the external station/stop ID for which the arrivals shall be retrieved. Required if id is not present. Such ID can be retrieved from the `stop_by_name` or `stop_by_coords`. Defaults to None.
        * direction (str, optional): If only vehicles departing or arriving from a certain direction shall be returned, specify the direction by giving the station/stop ID of the last stop on the journey. Defaults to None.
        * date (Union[str, datetime], optional): Sets the start date for which the departures shall be retrieved. Represented in the format YYYY-MM-DD. By default the current server date is used. Defaults to None.
        * time (Union[str, datetime], optional): Sets the start time for which the departures shall be retrieved. Represented in the format hh:mm[:ss] in 24h nomenclature. Seconds will be ignored for requests. By default the current server time is used. Defaults to None.
        * duration (int, optional): Set the interval size in minutes. Defaults to 60.
        * maxJourneys (int, optional): Maximum number of journeys to be returned. If no value is defined or -1, all departing/arriving services within the duration sized period are returned. Defaults to -1.
        * products (int, optional): Decimal value defining the product classes to be included in the search. It represents a bitmask combining bit number of a product as defined in the HAFAS raw data. Defaults to None.
        * operators (Union[str, list], optional): Only journeys provided by the given operators are part of the result. To filter multiple operators, separate the codes by comma. If the operator should not be part of the be trip, negate it by putting ! in front of it. Example: Filter for operator A and B: `operators=[A,B]`. Defaults to None.
        * lines (Union[str, list], optional): Only journeys running the given line are part of the result. To filter multiple lines, provide a list or separate the codes by comma. If the line should not be part of the be trip, negate it by putting ! in front of it. Defaults to None.
        * filterEquiv (bool, optional): Use `boardType` instead. Enables/disables the filtering of equivalent marked stops. Defaults to True.
        * attributes (Union[str, list], optional): Filter boards by one or more attribute codes of a journey. Multiple attribute as a list or as a string separated by comma. If the attribute should not be part of the result, negate it by putting ! in front of it. Defaults to None.
        * platforms (Union[str, list], optional): Filter boards by platform. Multiple platforms provided as a list or as a string separated by comma. Defaults to None.
        * passlist (bool, optional): Include a list of all passed waystops. Defaults to False.
        * boardType (Literal["DEP", "DEP_EQUIVS", "DEP_MAST", "DEP_STATION"], optional): Set the station departure board type to be used. DEP: Departure board as configured in HAFAS; DEP_EQUIVS: Departure board with all journeys at any masts and equivalent stops; DEP_MAST: Departure board at mast; DEP_STATION: Departure board with all journeys at any masts of the requested station. Defaults to "DEP".

    ### Returns:
        * dict: Output from RMV. Dict will contain "errorCode" and "errorText" if exception occurs.
    """    

    if json:
        headers = {"Accept": "application/json"}
    else:
        headers = {"Accept": "application/xml"}

    payload = {}

    for var, val in locals().items():
        if str(var) == "date":
            if val != None:
                if isinstance(val, datetime):
                    payload[str(var)] = val.strftime("%Y-%m-%d")
                else:
                    payload[str(var)] = val
        elif str(var) == "time":
            if val != None:
                if isinstance(val, datetime):
                    payload[str(var)] = val.strftime("%H:%M")
                else:
                    payload[str(var)] = val
        elif str(var) == "duration":
            if val != None:
                if isinstance(val, timedelta):
                    payload[str(var)] = val.minutes
                else:
                    payload[str(var)] = val
        elif str(var) == "boardType":
            if val != None:
                payload["type"] = val.upper()
        elif str(var) not in ["json", "headers", "payload"]:
            if val != None:
                payload[str(var)] = val

    output = get("https://www.rmv.de/hapi/departureBoard", params=payload, headers=headers)

    # Exceptions checker moved to normal methods
    # and exceptions will no longer raise in raw ones.
    # find_exception(output.json())

    if json:
        return output.json()
    else:
        return xmlparse(output.content)


from typing import OrderedDict


def weekdays_bitmask(data: OrderedDict[str, bool]) -> str:
    """Convert ordered dict with weekdays to a bitmask.

    ### Args:
        * data (OrderedDict[str, bool]): OrderedDict formatted as follows: OrderedDict(Monday=bool, Tuesday=bool, Wednesday=bool, Thursday=bool, Friday=bool, Saturday=bool, Sunday=bool)

    ### Returns:
        * str: _description_
    """    
    output = ""
    if len(data) != 7:
        raise ValueError("OrderedDict must be formatted as follows: OrderedDict(Monday=bool, Tuesday=bool, Wednesday=bool, Thursday=bool, Friday=bool, Saturday=bool, Sunday=bool)")
    for day in data:
        if data[day]:
            output += "1"
        else:
            output += "0"
    return output


from datetime import datetime
from pyrmv.methods.journey_detail import journey_detail
from pyrmv.methods.stop_by_id import stop_by_id


class LineArrival():

    def __init__(self, data, access_id: str):
        self.journey = journey_detail(access_id, data["JourneyDetailRef"]["ref"])
        self.status = data["JourneyStatus"]
        # self.messages = []
        self.name = data["name"]
        self.type = data["type"]
        self.stop_name = data["stop"]
        self.stop_id = data["stopid"]
        self.stop_id_ext = data["stopExtId"]
        self.stop = stop_by_id(access_id, self.stop_id)
        self.time = datetime.strptime(data["time"], "%H:%M:%S")
        self.date = datetime.strptime(data["date"], "%Y-%m-%d")
        if ("rtTime" in data) and ("rtDate" in data):
            self.time_real_time = datetime.strptime(data["rtTime"], "%H:%M:%S")
            self.date_real_time = datetime.strptime(data["rtDate"], "%Y-%m-%d")
        self.reachable = data["reachable"]
        self.origin = data["origin"]

    def __str__(self) -> str:
        return f"{self.name} coming from {self.origin} at {self.time.time()} {self.date.date()}"

class LineDeparture():

    def __init__(self, data, access_id: str):
        self.journey = journey_detail(access_id, data["JourneyDetailRef"]["ref"])
        self.status = data["JourneyStatus"]
        # self.messages = []
        self.name = data["name"]
        self.type = data["type"]
        self.stop_name = data["stop"]
        self.stop_id = data["stopid"]
        self.stop_id_ext = data["stopExtId"]
        self.stop = stop_by_id(access_id, self.stop_id)
        self.time = datetime.strptime(data["time"], "%H:%M:%S")
        self.date = datetime.strptime(data["date"], "%Y-%m-%d")
        if ("rtTime" in data) and ("rtDate" in data):
            self.time_real_time = datetime.strptime(data["rtTime"], "%H:%M:%S")
            self.date_real_time = datetime.strptime(data["rtDate"], "%Y-%m-%d")
        self.reachable = data["reachable"]
        self.direction = data["direction"]
        self.direction_flag = data["directionFlag"]
        
    def __str__(self) -> str:
        return f"{self.name} heading {self.direction} at {self.time.time()} {self.date.date()}"
            
class BoardArrival(list):

    def __init__(self, data: dict, access_id: str):
        super().__init__([])
        for line in data["Arrival"]:
            self.append(LineArrival(line, access_id))

    def __str__(self) -> str:
        lines = []
        for line in self:
            lines.append(str(line))
        return "Arrival board\n" + "\n".join(lines)

class BoardDeparture(list):

    def __init__(self, data: dict, access_id: str):
        super().__init__([])
        for line in data["Departure"]:
            self.append(LineDeparture(line, access_id))

    def __str__(self) -> str:
        lines = []
        for line in self:
            lines.append(str(line))
        return "Departure board\n" + "\n".join(lines)


class NotReadyYetError(Exception):
    """
    This method is not finished yet.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__


from pyrmv.errors.api_errors import ApiAuthError, ApiFormatError, ApiParamError, ApiQuotaError, ApiTooManyRequests
from pyrmv.errors.int_errors import IntError, IntGatewayError, IntTimeoutError
from pyrmv.errors.ps_errors import PsIncorrectParamError
from pyrmv.errors.sot_errors import SotAllTrainsFilteredError, SotAlreadyArrivedError, SotCancelledError, SotNotStartedError, SotStayOnTripError
from pyrmv.errors.svc_errors import SvcContextError, SvcDatetimeError, SvcDatetimePeriodError, SvcLocationArrivalError, SvcLocationDepartureError, SvcLocationEqualError, SvcLocationError, SvcLocationNearError, SvcLocationViaError, SvcNoMatchError, SvcNoResultError, SvcProductError, SvcSearchError
from pyrmv.errors.unknown_error import UnknownError

def find_exception(data: dict):
    """Scan returned dict for errorCode from RMV.
    Raises different exceptions if errorCode is not None. 

    ### Args:
        * data (dict): Response from RMV as a dict.

    ### Raises:
        * Any: Formatted as "errorCode -> errorText" if ApiParamError and UnknownError or as a single massage for others.
    """    
    if "errorCode" in data:

        if data["errorCode"] == "API_AUTH":
            raise ApiAuthError()

        elif data["errorCode"] == "API_QUOTA":
            raise ApiQuotaError()

        elif data["errorCode"] == "API_TOO_MANY_REQUESTS":
            raise ApiTooManyRequests()
            
        elif data["errorCode"] == "API_PARAM":
            raise ApiParamError(errorCode=data["errorCode"], errorText=data["errorText"])

        elif data["errorCode"] == "API_FORMAT":
            raise ApiFormatError()

        elif data["errorCode"] == "SVC_LOC":
            raise SvcLocationError()

        elif data["errorCode"] == "SVC_LOC_ARR":
            raise SvcLocationArrivalError()

        elif data["errorCode"] == "SVC_LOC_DEP":
            raise SvcLocationDepartureError()

        elif data["errorCode"] == "SVC_LOC_VIA":
            raise SvcLocationViaError()

        elif data["errorCode"] == "SVC_LOC_EQUAL":
            raise SvcLocationEqualError()

        elif data["errorCode"] == "SVC_LOC_NEAR":
            raise SvcLocationNearError()

        elif data["errorCode"] == "SVC_DATATIME":
            raise SvcDatetimeError()

        elif data["errorCode"] == "SVC_DATATIME_PERIOD":
            raise SvcDatetimePeriodError()

        elif data["errorCode"] == "SVC_PROD":
            raise SvcProductError()

        elif data["errorCode"] == "SVC_CTX":
            raise SvcContextError()

        elif data["errorCode"] == "SVC_NO_RESULT":
            raise SvcNoResultError()

        elif data["errorCode"] == "SVC_FAILED_SEARCH":
            raise SvcSearchError()

        elif data["errorCode"] == "SVC_NO_MATCH":
            raise SvcNoMatchError()

        elif data["errorCode"] == "INT_ERR":
            raise IntError()

        elif data["errorCode"] == "INT_GATEWAY":
            raise IntGatewayError()

        elif data["errorCode"] == "INT_TIMEOUT":
            raise IntTimeoutError()

        elif data["errorCode"] == "SOT_AT_DEST":
            raise SotAlreadyArrivedError()

        elif data["errorCode"] == "SOT_BEFORE_START":
            raise SotNotStartedError()

        elif data["errorCode"] == "SOT_CANCELLED":
            raise SotCancelledError()

        elif data["errorCode"] == "SOT_ALL_TRAINS_FILTERED":
            raise SotAllTrainsFilteredError()

        elif data["errorCode"] == "SOT_STAY_IN_CURRENT_CONNECTION":
            raise SotStayOnTripError()

        elif data["errorCode"] == "PARTIALSEARCH_INCORRECT_PARAM":
            raise PsIncorrectParamError()

        else:
            raise UnknownError(errorCode=data["errorCode"], errorText=data["errorText"])


class UnknownError(Exception):
    """Exception raised but error is not known.

    ### Attributes:
        * errorCode: Client error code from HAFAS ReST Request Errors.
        * errorText: Description of an error occurred.
    """

    def __init__(self, errorCode: str, errorText: str):
        self.errorCode = errorCode
        self.errorText = errorText
        super().__init__(self.errorText)

    def __str__(self):
        return f'{self.errorCode} -> {self.errorText}'


from .product import Product
from .rt_mode import RealTimeMode
from .lang import Language
from .location_type import LocationType
from .selection_mode import SelectionMode
from .filter_mode import FilterMode
from .board_type import BoardArrivalType, BoardDepartureType


from pyrmv.classes.Leg import Leg
from pyrmv.classes.Stop import StopTrip
from isodate import parse_duration

class Trip():
    """Trip object."""    
    
    def __init__(self, data: dict):
        
        self.raw_data = data
        self.origin = StopTrip(data["Origin"])
        self.destination = StopTrip(data["Destination"])

        self.legs = []
        for leg in data["LegList"]["Leg"]:
            self.legs.append(Leg(leg))

        self.calculation = data["calculation"]
        self.index = data["idx"]
        self.id = data["tripId"]
        self.ctx_recon = data["ctxRecon"]
        self.duration = parse_duration(data["duration"])

        if "rtDuration" in data:
            self.real_time_duration = parse_duration(data["rtDuration"])
        else:
            self.real_time_duration = None

        self.checksum = data["checksum"]
        
        if "transferCount" in data:
            self.transfer_count = data["transferCount"]
        else:
            self.transfer_count = 0

    def __str__(self) -> str:
        return f"Trip from {self.origin.name} to {self.destination.name} lasting {self.duration} ({self.real_time_duration}) with {len(self.legs)} legs and {self.transfer_count} transfers"


from .board_arrival import board_arrival
from .board_departure import board_departure
from .him_search import him_search
from .journey_detail import journey_detail
from .stop_by_coords import stop_by_coords
from .stop_by_name import stop_by_name
from .trip_find import trip_find
from .trip_recon import trip_recon


from isodate import parse_duration

class Gis():
    """Gis object."""    
    
    def __init__(self, ref: str, route: dict):
        
        self.ref = ref
        self.dist = route["dist"]
        self.duration = parse_duration(route["durS"])
        self.geo = route["dirGeo"]


from typing import List, Union
from pyrmv.classes.Stop import Stop
from pyrmv.enums.location_type import LocationType
from pyrmv.enums.lang import Language
from pyrmv.enums.selection_mode import SelectionMode
from pyrmv.enums.filter_mode import FilterMode
from pyrmv.raw.stop_by_name import stop_by_name as raw_stop_by_name
from pyrmv.utility.find_exception import find_exception

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

def stop_by_name(
    
        access_id: str,
        query: str,

        lang: Literal[Language.DE, Language.DA, Language.EN, Language.ES, Language.FR, Language.HU, Language.IT, Language.NL, Language.NO, Language.PL, Language.SV, Language.TR] = Language.EN,
        max_number: int = 10,
        stop_type: Literal[LocationType.A, LocationType.ALL, LocationType.AP, LocationType.P, LocationType.S, LocationType.SA, LocationType.SP] = LocationType.ALL,
        selection_mode: Literal[SelectionMode.SLCT_A, SelectionMode.SLCT_N] = None, # type: ignore
        coord_lat: Union[str, float] = None, # type: ignore
        coord_lon: Union[str, float] = None, # type: ignore
        radius: Union[int, float] = 1000,
        refine_id: str = None, # type: ignore
        stations: Union[str, list] = None, # type: ignore
        filter_mode: Literal[FilterMode.DIST_PERI, FilterMode.EXCL_PERI, FilterMode.SLCT_PERI] = FilterMode.DIST_PERI
    ) -> List[Stop]:
    """Method can be used to perform a pattern matching of a user input and to retrieve a list
    of possible matches in the journey planner database. Possible matches might be stops/stations,
    points of interest and addresses.  

    More detailed request is available as `raw.stop_by_name()`, however returns `dict` instead of `List[Stop]`.  

    ### Args:
        * access_id (`str`): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * query (`str`): Search for that token.
        * lang (`Literal[Language.DE, Language.DA, Language.EN, Language.ES, Language.FR, Language.HU, Language.IT, Language.NL, Language.NO, Language.PL, Language.SV, Language.TR]`, **optional**): The language of response. Defaults to `Language.EN`.
        * max_number (`int`, **optional**): Maximum number of returned stops. In range 1-1000. Defaults to `10`.
        * stop_type (`Literal[LocationType.A, LocationType.ALL, LocationType.AP, LocationType.P, LocationType.S, LocationType.SA, LocationType.SP]`, **optional**): Type filter for location types. Defaults to `LocationType.ALL`.
        * selection_mode (`Literal[SelectionMode.SLCT_A, SelectionMode.SLCT_N]`, **optional**): Selection mode for locations. `SelectionMode.SLCT_N`: Not selectable, `SelectionMode.SLCT_A`: Selectable. Defaults to `None`.
        * coord_lat (`Union[str, float]`, **optional**): Latitude of centre coordinate. Defaults to `None`.
        * coord_lon (`Union[str, float]`, **optional**): Longitude of centre coordinate. Defaults to `None`.
        * radius (`Union[int, float]`, **optional**): Search radius in meter around the given coordinate if any. Defaults to `1000`.
        * refine_id (`str`, **optional**): In case of an refinable location, this value takes the ID of the refinable one of a previous result. Defaults to `None`.
        * stations (`Union[str, list]`, **optional**): Filter for stations. Matches if the given value is prefix of any station in the itinerary. As a list or as a string separated by comma. Defaults to `None`.
        * filter_mode (`Literal[FilterMode.DIST_PERI, FilterMode.EXCL_PERI, FilterMode.SLCT_PERI]`, **optional**): Filter modes for nearby searches. Defaults to `FilterMode.DIST_PERI`.

    ### Returns:
        * List[Stop]: List of Stop objects. Empty list if none found.
    """    

    if selection_mode == None:
        selection_mode = None
    else:
        selection_mode = selection_mode.code

    stops = []
    stops_raw = raw_stop_by_name(
        accessId=access_id,
        inputString=query,
        lang=lang.code,
        maxNo=max_number,
        stopType=stop_type.code,
        locationSelectionMode=selection_mode,
        coordLat=coord_lat,
        coordLong=coord_lon,
        radius=radius,
        refineId=refine_id,
        stations=stations,
        filterMode=filter_mode.code
    )

    find_exception(stops_raw)

    for stop in stops_raw["stopLocationOrCoordLocation"]:
        if "StopLocation" in stop:
            stops.append(Stop(stop["StopLocation"]))
        elif "CoordLocation" in stop:
            stops.append(Stop(stop["CoordLocation"]))

    return stops


# Class is taken from source code of Pyrogram:
# https://github.com/pyrogram/pyrogram/blob/master/pyrogram/enums/auto_name.py

from enum import Enum
from pyrmv.const.product import PRODUCTS

class AutoName(Enum):
    def __init__(self, code) -> None:
        self.code = code

    def _generate_next_value_(self, *args):
        return self.lower()

    def __repr__(self):
        return f"pyrmv.enums.{self}"

class AutoNameProduct(AutoName):
    def __init__(self, code) -> None:
        self.code = PRODUCTS[code]


from typing import List, Union
from pyrmv.classes.Stop import Stop
from pyrmv.enums.location_type import LocationType
from pyrmv.enums.lang import Language
from pyrmv.enums.selection_mode import SelectionMode
from pyrmv.raw.stop_by_coords import stop_by_coords as raw_stop_by_coords
from pyrmv.utility.find_exception import find_exception

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

def stop_by_coords(
    
        access_id: str,
        coords_lat: Union[str, float],
        coords_lon: Union[str, float],

        lang: Literal[Language.DE, Language.DA, Language.EN, Language.ES, Language.FR, Language.HU, Language.IT, Language.NL, Language.NO, Language.PL, Language.SV, Language.TR] = Language.EN,
        radius: Union[int, float] = 1000,
        max_number: int = 10,
        stop_type: Literal[LocationType.S, LocationType.P, LocationType.SP, LocationType.SE, LocationType.SPE] = LocationType.S,
        selection_mode: Literal[SelectionMode.SLCT_A, SelectionMode.SLCT_N] = None, # type: ignore
    ) -> List[Stop]:
    """Method returns a list of stops around a given center coordinate.
    The returned results are ordered by their distance to the center coordinate.  

    More detailed request is available as `raw.stop_by_coords()`, however returns `dict` instead of `List[Stop]`.

    ### Args:
        * access_id (`str`): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * coords_lat (`Union[str, float]`): Latitude of centre coordinate.
        * coords_lon (`Union[str, float]`): Longitude of centre coordinate.
        * lang (`Literal[Language.DE, Language.DA, Language.EN, Language.ES, Language.FR, Language.HU, Language.IT, Language.NL, Language.NO, Language.PL, Language.SV, Language.TR]`, **optional**): The language of response. Defaults to `Language.EN`.
        * radius (`Union[int, float]`, **optional**): Search radius in meter around the given coordinate if any. Defaults to `1000`.
        * max_number (`int`, **optional**): Maximum number of returned stops. Defaults to `10`.
        * stop_type (`Literal[LocationType.S, LocationType.P, LocationType.SP, LocationType.SE, LocationType.SPE]`, **optional**): Type filter for location types. Defaults to `LocationType.S`.
        * selection_mode (`Literal[SelectionMode.SLCT_A, SelectionMode.SLCT_N]`, **optional**): Selection mode for locations. `SelectionMode.SLCT_N`: Not selectable, `SelectionMode.SLCT_A`: Selectable. Defaults to `None`.

    ### Returns:
        * List[Stop]: List of Stop objects. Empty list if none found.
    """    

    if selection_mode == None:
        selection_mode = None
    else:
        selection_mode = selection_mode.code

    stops = []
    stops_raw = raw_stop_by_coords(
        accessId=access_id,
        originCoordLat=coords_lat,
        originCoordLong=coords_lon,
        lang=lang.code,
        radius=radius,
        maxNo=max_number,
        stopType=stop_type.code,
        locationSelectionMode=selection_mode
    )

    find_exception(stops_raw)

    for stop in stops_raw["stopLocationOrCoordLocation"]:
        if "StopLocation" in stop:
            stops.append(Stop(stop["StopLocation"]))
        elif "CoordLocation" in stop:
            stops.append(Stop(stop["CoordLocation"]))

    return stops


class Ticket():
    """
    {
        "externalContent": {
            "content": "eyJSTVZUaWNrZXQiOnsiVGV4dExvbmciOiJFaW56ZWxmYWhydCBLaW5kZXIiLCJUZXh0TWlkZGxlIjoiRWluemVsZmFocnQgS2luZGVyIiwiVGV4dFNob3J0IjoiRWluemVsZmFocnQgS2luZGVyIiwiUmVtYXJrIjoiUmVnZWxwcmVpcyIsIlZhbGlkRnJvbSI6IjI0LjA5LjIwMjIiLCJUYXJpZmZMZXZlbFByaW50IjoiNDAiLCJUYXJpZmZMZXZlbEludGVybiI6IjYwIiwiUHJpY2UiOiI0NDUiLCJUWkNvZGVTdGFydFJlcSI6IjUyMzMiLCJUWkNvZGVTdGFydFJlcyI6IjUyMzMiLCJUWk5hbWVTdGFydCI6IlNjaG1pdHRlbiBLaXJjaGUiLCJWaWFJZCI6IjAiLCJWaWFUZXh0IjoiS8O2bmlnc3RlaW4vSG9oZW1hcmsiLCJUWkNvZGVEZXN0UmVxIjoiNTAwMSIsIlRaQ29kZURlc3RSZXMiOiI1MDAxIiwiVFpOYW1lRGVzdCI6IkZyYW5rZnVydCAoTWFpbikgSGF1cHR3YWNoZSIsIlR5cGVLZXkiOiIzMzAwIiwiU3VidmVudGlvbiI6IjAiLCJUYXJpZmZWZXJzaW9uIjoiNDEiLCJLYXJ0ZW50eXAiOiJLVFlQLUVpbnplbGZhaHJrYXJ0ZSIsIkthcnRlbnR5cEFuemVpZ2VuYW1lIjoiRWluemVsZmFocnQiLCJLQVByb2R1a3RudW1tZXIiOiIxMDE4NiIsIktBSW5mb1RleHQiOiIiLCJWZXJhbnR3b3J0bGljaGVyIjoiUk1WIiwiVmVyYnVlbmRlIjoiUk1WIiwiRXhwb3J0QXR0cmlidXRlIjoiQVRUUi1WVFJCLUVpbnplbHByZWlzPUtpbmR8QVRUUi1WVFJCLXJlcGxTb3J0ZW49MjIwMCAyNTAwIDI4MDAgMzgwMHxBVFRSLUtUWVAtYmlsbGlnVHI9MSIsIkZyZWlnYWJlSWQiOiIyMDczMCIsIkZyZWlnYWJlIjoiMjY0NSwgMzY3MCwgNTAwMSwgNTAwMiwgNTAwMywgNTAwNCwgNTAwNSwgNTAwNiwgNTAxMSwgNTAxMywgNTAxNSwgNTAyMSwgNTAyMiwgNTAyMywgNTAzMSwgNTA0MSwgNTA0MiwgNTA0NiwgNTA1OSwgNTEwMCwgNTIzMywgNjIzOSwgNjYyNiwgNjYzNiwgNjYzNywgNjYzOSwgNjY0OCwgNjY0OSwgNjY1MCwgNjY1NSwgNjY2MCwgNjY2NCwgNjY5MSIsIkZyZWlnYWJlT3ZlcnZpZXciOiJCYWQgSG9tYnVyZyB2LmQuSMO2aGU7QmFkIFNvZGVuIGEuVHMuO0VzY2hib3JuO0ZyaWVkcmljaHNkb3JmO0dsYXNow7x0dGVuIn19",
            "contentType": "LIB_TARIFF"
        },
        "name": "Einzelfahrkarte Regelpreis",
        "desc": "Kinder",
        "price": 445,
        "cur": "EUR",
        "shpCtx": "{\"RMVTicket\":{\"TextLong\":\"Einzelfahrt Kinder\",\"TextMiddle\":\"Einzelfahrt Kinder\",\"TextShort\":\"Einzelfahrt Kinder\",\"Remark\":\"Regelpreis\",\"ValidFrom\":\"24.09.2022\",\"TariffLevelPrint\":\"40\",\"TariffLevelIntern\":\"60\",\"Price\":\"445\",\"TZCodeStartReq\":\"5233\",\"TZCodeStartRes\":\"5233\",\"TZNameStart\":\"Schmitten Kirche\",\"ViaId\":\"0\",\"ViaText\":\"Königstein/Hohemark\",\"TZCodeDestReq\":\"5001\",\"TZCodeDestRes\":\"5001\",\"TZNameDest\":\"Frankfurt (Main) Hauptwache\",\"TypeKey\":\"3300\",\"Subvention\":\"0\",\"TariffVersion\":\"41\",\"Kartentyp\":\"KTYP-Einzelfahrkarte\",\"KartentypAnzeigename\":\"Einzelfahrt\",\"KAProduktnummer\":\"10186\",\"KAInfoText\":\"\",\"Verantwortlicher\":\"RMV\",\"Verbuende\":\"RMV\",\"ExportAttribute\":\"ATTR-VTRB-Einzelpreis=Kind|ATTR-VTRB-replSorten=2200 2500 2800 3800|ATTR-KTYP-billigTr=1\",\"FreigabeId\":\"20730\",\"Freigabe\":\"2645, 3670, 5001, 5002, 5003, 5004, 5005, 5006, 5011, 5013, 5015, 5021, 5022, 5023, 5031, 5041, 5042, 5046, 5059, 5100, 5233, 6239, 6626, 6636, 6637, 6639, 6648, 6649, 6650, 6655, 6660, 6664, 6691\",\"FreigabeOverview\":\"Bad Homburg v.d.Höhe;Bad Soden a.Ts.;Eschborn;Friedrichsdorf;Glashütten\"}}"
    }
    ###############################################
    "shpctx": {
        "RMVTicket":
            {
                "TextLong": "Einzelfahrt Kinder",
                "TextMiddle": "Einzelfahrt Kinder",
                "TextShort": "Einzelfahrt Kinder",
                "Remark": "Regelpreis",
                "ValidFrom": "24.09.2022",
                "TariffLevelPrint": "40",
                "TariffLevelIntern": "60",
                "Price": "445",
                "TZCodeStartReq": "5233",
                "TZCodeStartRes": "5233",
                "TZNameStart": "Schmitten Kirche",
                "ViaId": "0",
                "ViaText": "Königstein/Hohemark",
                "TZCodeDestReq": "5001",
                "TZCodeDestRes": "5001",
                "TZNameDest": "Frankfurt (Main) Hauptwache",
                "TypeKey": "3300",
                "Subvention": "0",
                "TariffVersion": "41",
                "Kartentyp": "KTYP-Einzelfahrkarte",
                "KartentypAnzeigename": "Einzelfahrt",
                "KAProduktnummer": "10186",
                "KAInfoText": "",
                "Verantwortlicher": "RMV",
                "Verbuende": "RMV",
                "ExportAttribute": "ATTR-VTRB-Einzelpreis=Kind|ATTR-VTRB-replSorten=2200 2500 2800 3800|ATTR-KTYP-billigTr=1",
                "FreigabeId": "20730",
                "Freigabe":"2645, 3670, 5001, 5002, 5003, 5004, 5005, 5006, 5011, 5013, 5015, 5021, 5022, 5023, 5031, 5041, 5042, 5046, 5059, 5100, 5233, 6239, 6626, 6636, 6637, 6639, 6648, 6649, 6650, 6655, 6660, 6664, 6691",
                "FreigabeOverview": "Bad Homburg v.d.Höhe;Bad Soden a.Ts.;Eschborn;Friedrichsdorf;Glashütten"
        }        
    }
    """

    def __init__(self) -> None:
        pass


from datetime import datetime
from typing import Union
from pyrmv.classes.Journey import Journey
from pyrmv.enums.rt_mode import RealTimeMode
from pyrmv.raw.journey_detail import journey_detail as raw_journey_detail
from pyrmv.utility.find_exception import find_exception

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

def journey_detail(

        access_id: str,
        id: str,
        date: Union[str, datetime] = None, # type: ignore
        real_time_mode: Literal[RealTimeMode.FULL, RealTimeMode.INFOS, RealTimeMode.OFF, RealTimeMode.REALTIME, RealTimeMode.SERVER_DEFAULT] = None, # type: ignore
        from_id: str = None, # type: ignore
        from_index: int = None, # type: ignore
        to_id: str = None, # type: ignore
        to_index: int = None # type: ignore
    ) -> Journey:
    """The journey_detail method will deliver information about the complete route of a vehicle. The journey
    identifier is part of a trip or `board_departure()` response. It contains a list of all stops/stations of this journey
    including all departure and arrival times (with real-time data if available) and additional information like
    specific attributes about facilities and other texts.  

    More detailed request is available as `raw.journey_detail()`, however returns `dict` instead of `Journey`.

    ### Args:
        * access_id (`str`): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * id (`str`): Specifies the internal journey id of the journey shall be retrieved. Maximum length 512.
        * date (`Union[str, datetime]`, **optional**): Day of operation. Represented in the format `YYYY-MM-DD` or as a datetime object. By default the current server date is used. Defaults to `None`.
        * real_time_mode (`Literal[RealTimeMode.FULL, RealTimeMode.INFOS, RealTimeMode.OFF, RealTimeMode.REALTIME, RealTimeMode.SERVER_DEFAULT]`, **optional**): Set the realtime mode to be used. Read more about this in HAFAS ReST Documentation. Defaults to `None`.
        * from_id (`str`, **optional**): Specifies the station/stop ID the partial itinerary shall start from. Defaults to `None`.
        * from_index (`int`, **optional**): Specifies the station/stop index the partial itinerary shall start from. Defaults to `None`.
        * to_id (`str`, **optional**): Specifies the station/stop ID the partial itinerary shall end at. Defaults to `None`.
        * to_index (`int`, **optional**): Specifies the station/stop index the partial itinerary shall end at. Defaults to `None`.

    ### Returns:
        * Journey: Instance of Journey object.
    """    

    if real_time_mode == None:
        real_time_mode = None
    else:
        real_time_mode = real_time_mode.code

    journey_raw = raw_journey_detail(
        accessId=access_id,
        id=id,
        date=date,
        rtMode=real_time_mode,
        fromId=from_id,
        fromIdx=from_index,
        toId=to_id,
        toIdx=to_index
    )

    find_exception(journey_raw)

    return Journey(journey_raw)


from datetime import datetime
from requests import get
from typing import Union
from xmltodict import parse as xmlparse

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# 2.17. Reconstruction (recon)
def trip_recon(accessId: str,
        ctx: str,
        json: bool = True,
        poly: bool = False,
        polyEnc: Literal["DLT", "GPA", "N"] = "N",
        date: Union[str, datetime] = None, # type: ignore
        useCombinedComparison: bool = None, # type: ignore
        acceptGaps: bool = None, # type: ignore
        allowDummySections: bool = None, # type: ignore
        flagAllNonReachable: bool = None, # type: ignore
        matchCatStrict: bool = None, # type: ignore
        matchIdNonBlank: bool = None, # type: ignore
        matchIdStrict: bool = None, # type: ignore
        matchNumStrict: bool = None, # type: ignore
        matchRtType: bool = None, # type: ignore
        enableRtFullSearch: bool = None, # type: ignore
        enableReplacements: bool = None, # type: ignore
        arrL: int = None, # type: ignore
        arrU: int = None, # type: ignore
        depL: int = None, # type: ignore
        depU: int = None, # type: ignore
        passlist: bool = False,
        showPassingPoints: bool = False,
        rtMode: Literal["FULL", "INFOS", "OFF", "REALTIME", "SERVER_DEFAULT"] = None, # type: ignore
        eco: bool = False,
        ecoCmp: bool = False,
        ecoParams: str = None, # type: ignore
        tariff: bool = None, # type: ignore
        trafficMessages: bool = None, # type: ignore
        travellerProfileData: str = None # type: ignore
    ) -> dict:
    """Reconstructing a trip can be achieved using the resconstruction context provided by any trip result in the
    ctxRecon attribute of Trip element. The result will be a true copy of the original trip search result given
    that the underlying data did not change.  

    Read more about this in section 2.17. "Reconstruction (recon)" of HAFAS ReST Documentation.  

    ### Args:
        * accessId (str): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * ctx (str): Specifies the reconstruction context.
        * json (bool, optional): Whether response should be retrieved as JSON. XML is returned if False. Defaults to True.
        * poly (bool, optional): Enables/disables the calculation of the polyline for each leg of the trip except any GIS route. Defaults to False.
        * polyEnc (Literal["DLT", "GPA", "N"], optional): Defines encoding of the returned polyline. Possible values are "N" (no encoding / compression), "DLT" (delta to the previous coordinate), "GPA" (Google encoded polyline format). Defaults to "N".
        * date (Union[str, datetime], optional): Sets the start date for which the departures shall be retrieved. Represented in the format `YYYY-MM-DD`. This parameter will force the ser-vice to reconstruct the trip on that specific date. If the trip is not available on that date, because it does not operate, the error code SVC_NO_RESULT will be returned. Defaults to None.
        * useCombinedComparison (bool, optional): Compare based on combined output name - `False`: Compare parameters (category, line, train number) individually. Defaults to None.
        * acceptGaps (bool, optional): Accept an incomplete description of the connection (with gaps) i.e. missing walks/transfers. Defaults to None.
        * allowDummySections (bool, optional): Allow a partial reconstruction that will not lead to a reconstruction failure if sections are not reconstructable. Instead, for theses inconstructable sections, dummy sections will be created in the result. Defaults to None.
        * flagAllNonReachable (bool, optional): Should all non-reachable journeys be flagged (`True`), or only the first one encountered? Defaults to None.
        * matchCatStrict (bool, optional): Should the category (Gattung) match exactly? Only applicable if `useCombinedComparison` is `False`. Defaults to None.
        * matchIdNonBlank (bool, optional): Should the train identifier (Zugbezeichner) without whitespace match? Defaults to None.
        * matchIdStrict (bool, optional): Should the train identifier (Zugbezeichner) match exactly? Defaults to None.
        * matchNumStrict (bool, optional): Should the train number (Zugnummer) match exactly? Only applicable if `useCombinedComparison` is `False`. Defaults to None.
        * matchRtType (bool, optional): Should the realtime type that journeys are based on (e.g. SOLL, IST, additional, deviation, …) be considered? Defaults to None.
        * enableRtFullSearch (bool, optional): By default, the reconstruction request makes one attempt for each journey within the scheduled data. However, the scheduled data may not necessarily reflect basic realtime properties of the journeys therein. In such a case, one may enable a two-step approach which we call "full search", i.e. search for matching journeys in the scheduled data in a first step. If this fails, then search for matching journeys in the realtime data. Defaults to None.
        * enableReplacements (bool, optional): If set to true replaces cancelled journeys with their replacement journeys if possible. Defaults to None.
        * arrL (int, optional): Lower deviation in minutes within interval [0, 720] indicating "how much earlier than original arrival". Defaults to None.
        * arrU (int, optional): Upper deviation in minutes within interval [0, 720] indicating "how much later than original arrival". Defaults to None.
        * depL (int, optional): Lower deviation in minutes within interval [0, 720] indicating "how much earlier than original departure". Defaults to None.
        * depU (int, optional): Upper deviation in minutes within interval [0, 720] indicating "how much later than original departure". Defaults to None.
        * passlist (bool, optional): Enables/disables the return of the passlist for each leg of the trip. Defaults to False.
        * showPassingPoints (bool, optional): Enables/disables the return of stops having no alighting and boarding in its passlist for each leg of the trip. Needs passlist parameter enabled. Defaults to False.
        * rtMode (Literal["FULL", "INFOS", "OFF", "REALTIME", "SERVER_DEFAULT"], optional): Set the realtime mode to be used. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * eco (bool, optional): Enables/disables eco value calculation. Defaults to False.
        * ecoCmp (bool, optional): Enables/disables eco comparison. Defaults to False.
        * ecoParams (str, optional): Provide additional eco parameters. Values vary. Defaults to None.
        * tariff (bool, optional): Enables/disables the output of tariff data. The default is configurable via provisioning. Defaults to None.
        * trafficMessages (bool, optional): Enables/disables the output of traffic messages. The default is configurable via provisioning. Defaults to None.
        * travellerProfileData (str, optional): Traveller profile data. Structure depends on set up. Defaults to None.

    ### Returns:
        * dict: Output from RMV. Dict will contain "errorCode" and "errorText" if exception occurs.
    """    

    if json:
        headers = {"Accept": "application/json"}
    else:
        headers = {"Accept": "application/xml"}

    payload = {}

    for var, val in locals().items():
        if str(var) == "date":
            if val != None:
                if isinstance(val, datetime):
                    payload[str(var)] = val.strftime("%Y-%m-%d")
                else:
                    payload[str(var)] = val
        elif str(var) not in ["json", "headers", "payload"]:
            if val != None:
                payload[str(var)] = val

    output = get("https://www.rmv.de/hapi/recon", params=payload, headers=headers)

    # Exceptions checker moved to normal methods
    # and exceptions will no longer raise in raw ones.
    # find_exception(output.json())

    if json:
        return output.json()
    else:
        return xmlparse(output.content)


from enum import auto
from .auto_name import AutoName

class LocationType(AutoName):
    """Enumeration used to declare types of location filter.  
    
    * S - Search for station/stops only
    * A - Search for addresses only
    * P - Search for POIs only
    * AP - Search for addresses and POIs
    * SA - Search for station/stops and addresses
    * SE - Search for stations/stops and entrypoints
    * SP - Search for stations/stops and POIs
    * ALL - Search in all existing location pools
    * SPE - Search for stations/stops, POIs and entrypoints

    Note that not all location types may be available for a certain methods.
    Make sure that selected type is supported by method by looking up 
    its acceptable types in argument description.
    """

    S = auto()
    "Search for station/stops only"

    A = auto()
    "Search for addresses only"

    P = auto()
    "Search for POIs only"

    AP = auto()
    "Search for addresses and POIs"

    SA = auto()
    "Search for station/stops and addresses"

    SE = auto()
    "Search for stations/stops and entrypoints"

    SP = auto()
    "Search for stations/stops and POIs"

    ALL = auto()
    "Search in all existing location pools"

    SPE = auto()
    "Search for stations/stops, POIs and entrypoints"


class PsIncorrectParamError(Exception):
    """
    An invalid parameter combination was requested, i.e. the
    defined range of stable segments encompassed all public
    transport sections or it was attempted to search
    forward/backward from the end/beginning of the
    connection.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__


from setuptools import setup

setup(
    name="pyrmv",
    version="0.2.9",
    author="Profitroll",
    description="Small module that makes your journey with RMV REST API somehow easier.",
    long_description="## PythonRMV\n\nSmall module that makes your journey with RMV REST API somehow easier. Based fully on official RMV API reference and HAFAS documentation.\n\n# Usage\n\n```py\nimport pyrmv\n\n# Set API key\naccess_id = \"Something\"\n\n# Get origin's and destination's location\norigin = pyrmv.stop_by_name(access_id, \"Frankfurt Hauptbahnhof\", max_number=3)[0]\ndestination = pyrmv.stop_by_coords(access_id, 50.099613, 8.685449, max_number=3)[0]\n\n# Find a trip by locations got\ntrip = pyrmv.trip_find(access_id, origin_id=origin.id, dest_id=destination.id)\n```\n\n# Frequently Asked Questions\n\n- [Why are there raw versions and formatted ones?](#why-are-there-raw-versions-and-formatted-ones)\n- [Some methods work slightly different](#some-methods-work-slightly-different)\n- [Documentation is not perfectly clear](#documentation-is-not-perfectly-clear)\n\n## Why are there raw versions and formatted ones?\n\nFor the purposes of my projects I don't really need all the stuff RMV gives (even though it's not much).\nI only need some specific things. However I do understand that in some cases other users may find\nthose methods quite useful so I implemented them as well.\n\n\n## Some methods work slightly different\n\nCan be. Not all function arguments written may work perfectly because I simply did not test each and\nevery request. Some of arguments may be irrelevant in my use-case and the others are used quite rare at all.\nJust [make an issue](https://git.end-play.xyz/profitroll/PythonRMV/issues/new) and I'll implement it correct when I'll have some free time.\n\n## Documentation is not perfectly clear\n\nOf course docs cannot be perfect as a python docstring, especially if sometimes I don't\nknow how things should correctly work. That's why you get HAFAS API docs in addition to your\nRMV API key. Just use my functions together with those docs, if you want to build something\nreally sophisticated. However I'm not sure whether RMV supports that many HAFAS features publicly.\n\n# To-Do\n## General\n- [ ] Docs in Wiki\n\n## Raw methods\n- [x] arrivalBoard (board_arrival)  \n- [x] departureBoard (board_departure)  \n- [x] himsearch (him_search)  \n- [x] journeyDetail (journey_detail)\n- [x] location.nearbystops (stop_by_coords)  \n- [x] location.name (stop_by_name)  \n- [x] trip (trip_find)  \n- [x] recon (trip_recon)\n\n## Normal methods\n- [x] arrivalBoard (board_arrival)  \n- [x] departureBoard (board_departure)  \n- [ ] himsearch (him_search)  \n- [x] journeyDetail (journey_detail)\n- [x] location.nearbystops (stop_by_coords)  \n- [x] location.name (stop_by_name)  \n- [x] trip (trip_find)  \n- [ ] recon (trip_recon)",
    long_description_content_type="text/markdown",
    author_email="profitroll@end-play.xyz",
    url="https://git.end-play.xyz/profitroll/PythonRMV",
    project_urls={
        "Bug Tracker": "https://git.end-play.xyz/profitroll/PythonRMV/issues",
        "Documentation": "https://git.end-play.xyz/profitroll/PythonRMV/wiki",
        "Source Code": "https://git.end-play.xyz/profitroll/PythonRMV.git",
    },
    packages=[
        "pyrmv",
        "pyrmv.raw",
        "pyrmv.const",
        "pyrmv.enums",
        "pyrmv.errors",
        "pyrmv.methods",
        "pyrmv.utility",
        "pyrmv.classes"
    ],
    install_requires=[
        "requests",
        "xmltodict",
        "isodate"
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities"
    ]
)


from enum import auto
from .auto_name import AutoName

class BoardArrivalType(AutoName):
    """Enumeration used to declare types of arrival board.  
    
    * ARR - Arrival board as configured in HAFAS
    * ARR_EQUIVS - Arrival board with all journeys at any masts and equivalent stops
    * ARR_MAST - Arrival board at mast
    * ARR_STATION - Arrival board with all journeys at any masts of the requested station
    """

    ARR = auto()
    "Arrival board as configured in HAFAS"

    ARR_EQUIVS = auto()
    "Arrival board with all journeys at any masts and equivalent stops"

    ARR_MAST = auto()
    "Arrival board at mast"

    ARR_STATION = auto()
    "Arrival board with all journeys at any masts of the requested station"

class BoardDepartureType(AutoName):
    """Enumeration used to declare types of departure board.  
    
    * DEP - Departure board as configured in HAFAS
    * DEP_EQUIVS - Departure board with all journeys at any masts and equivalent stops
    * DEP_MAST - Departure board at mast
    * DEP_STATION - Departure board with all journeys at any masts of the requested station
    """

    DEP = auto()
    "Departure board as configured in HAFAS"

    DEP_EQUIVS = auto()
    "Departure board with all journeys at any masts and equivalent stops"

    DEP_MAST = auto()
    "Departure board at mast"

    DEP_STATION = auto()
    "Departure board with all journeys at any masts of the requested station"


from datetime import datetime
from requests import get
from typing import Union
from xmltodict import parse as xmlparse

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# 2.26. Journey Detail (journeyDetail)
def journey_detail(accessId: str,
        id: str,
        json: bool = True,
        date: Union[str, datetime] = None, # type: ignore
        poly: bool = False,
        polyEnc: Literal["DLT", "GPA", "N"] = "N",
        showPassingPoints: bool = False,
        rtMode: Literal["FULL", "INFOS", "OFF", "REALTIME", "SERVER_DEFAULT"] = None, # type: ignore
        fromId: str = None, # type: ignore
        fromIdx: str = None, # type: ignore
        toId: str = None, # type: ignore
        toIdx: str = None, # type: ignore
        baim: bool = False
    ) -> dict:
    """The journey_detail method will deliver information about the complete route of a vehicle. The journey
    identifier is part of a trip or departureBoard response. It contains a list of all stops/stations of this journey
    including all departure and arrival times (with real-time data if available) and additional information like
    specific attributes about facilities and other texts.  

    Read more about this in section 2.26. "Journey Detail (journeyDetail)" of HAFAS ReST Documentation.  

    ### Args:
        * accessId (str): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * id (str): Specifies the internal journey id of the journey shall be retrieved. Maximum length 512.
        * json (bool, optional): Whether response should be retrieved as JSON. XML is returned if False. Defaults to True.
        * date (Union[str, datetime], optional): Day of operation. Represented in the format `YYYY-MM-DD`. By default the current server date is used. Defaults to None.
        * poly (bool, optional): Enables/disables the calculation of the polyline for each leg of the trip except any GIS route. Defaults to False.
        * polyEnc (Literal["DLT", "GPA", "N"], optional): Defines encoding of the returned polyline. Possible values are "N" (no encoding / compression), "DLT" (delta to the previous coordinate), "GPA" (Google encoded polyline format). Defaults to "N".
        * showPassingPoints (bool, optional): Enables/disables the return of stops having no alighting and no boarding in its passlist for each leg of the trip. Defaults to False.
        * rtMode (Literal["FULL", "INFOS", "OFF", "REALTIME", "SERVER_DEFAULT"], optional): Set the realtime mode to be used. Read more about this in HAFAS ReST Documentation. Defaults to None.
        * fromId (str, optional): Specifies the station/stop ID the partial itinerary shall start from. Defaults to None.
        * fromIdx (str, optional): Specifies the station/stop index the partial itinerary shall start from. Defaults to None.
        * toId (str, optional): Specifies the station/stop ID the partial itinerary shall end at. Defaults to None.
        * toIdx (str, optional): Specifies the station/stop index the partial itinerary shall end at. Defaults to None.
        * baim (bool, optional): Enables/disables BAIM search and response. Defaults to False.

    ### Returns:
        * dict: Output from RMV. Dict will contain "errorCode" and "errorText" if exception occurs.
    """    

    if json:
        headers = {"Accept": "application/json"}
    else:
        headers = {"Accept": "application/xml"}

    payload = {}

    for var, val in locals().items():
        if str(var) == "rtMode":
            if val != None:
                payload["rtMode"] = val.upper()
        elif str(var) not in ["json", "headers", "payload"]:
            if val != None:
                payload[str(var)] = val

    output = get("https://www.rmv.de/hapi/journeyDetail", params=payload, headers=headers)

    # Exceptions checker moved to normal methods
    # and exceptions will no longer raise in raw ones.
    # find_exception(output.json())

    if json:
        return output.json()
    else:
        return xmlparse(output.content)


from datetime import datetime
from typing import List, Union
from pyrmv.classes.Trip import Trip
from pyrmv.raw.trip_find import trip_find as raw_trip_find
from pyrmv.enums.rt_mode import RealTimeMode
from pyrmv.enums.lang import Language
from pyrmv.utility.find_exception import find_exception

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

def trip_find(
    
        access_id: str,
        lang: Literal[Language.DE, Language.DA, Language.EN, Language.ES, Language.FR, Language.HU, Language.IT, Language.NL, Language.NO, Language.PL, Language.SV, Language.TR] = Language.EN,

        origin_id: str = None, # type: ignore
        origin_id_ext: str = None, # type: ignore
        origin_coord_lat: Union[str, float] = None, # type: ignore
        origin_coord_lon: Union[str, float] = None, # type: ignore
        origin_coord_name: str = None, # type: ignore

        destination_id: str = None, # type: ignore
        destination_id_ext: str = None, # type: ignore
        destination_coord_lat: Union[str, float] = None, # type: ignore
        destination_coord_lon: Union[str, float] = None, # type: ignore
        destination_coord_name: str = None, # type: ignore

        via: str = None, # type: ignore
        via_id: str = None, # type: ignore
        via_gis: str = None, # type: ignore
        via_wait_time: int = 0,

        avoid: str = None, # type: ignore
        avoid_id: str = None, # type: ignore

        change_time_percent: int = 100,
        change_time_min: int = None, # type: ignore
        change_time_max: int = None, # type: ignore
        change_time_add: int = None, # type: ignore
        change_max: int = None, # type: ignore

        date: Union[str, datetime] = None, # type: ignore
        time: Union[str, datetime] = None, # type: ignore
        
        search_arrival: bool = False,

        trips_after_time: int = None, # type: ignore
        trips_before_time: int = None, # type: ignore

        context: str = None, # type: ignore

        passlist: bool = False,
        operators: Union[str, list] = None, # type: ignore
        
        lines: Union[str, list] = None, # type: ignore
        lineids: Union[str, list] = None, # type: ignore

        iv_include: bool = False,
        iv_only: bool = False,
        
        bike_carriage: bool = False,

        passing_points: bool = False,

        real_time_mode: Literal[RealTimeMode.FULL, RealTimeMode.INFOS, RealTimeMode.OFF, RealTimeMode.REALTIME, RealTimeMode.SERVER_DEFAULT] = None, # type: ignore

        include_earlier: bool = False,
        ict_alternatives: bool = False,
        tariff: bool = None, # type: ignore
        messages: bool = False,
        frequency: bool = True
    ) -> List[Trip]:
    """The trip service calculates a trip from a specified origin to a specified destination. These might be
    stop/station IDs or coordinates based on addresses and points of interest validated by the location service or
    coordinates freely defined by the client.

    More detailed request is available as `raw.trip_find()`, however returns `dict` instead of `List[Trip]`.  

    ### Args:
        * access_id (`str`): Access ID for identifying the requesting client. Get your key on [RMV website](https://opendata.rmv.de/site/start.html).
        * lang (`Literal[Language.DE, Language.DA, Language.EN, Language.ES, Language.FR, Language.HU, Language.IT, Language.NL, Language.NO, Language.PL, Language.SV, Language.TR]`, **optional**): The language of response. Defaults to `Language.EN`.
        * origin_id (`str`, **optional**): Specifies the station/stop ID of the origin for the trip. Such ID can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * origin_id_ext (`str`, **optional**): Deprecated. Please use originId as it supports external IDs. Specifies the external station/stop ID of the origin for the trip. Such ID can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * origin_coord_lat (`Union[str, float]`, **optional**): Latitude of station/stop coordinate of the trip's origin. The coordinate can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * origin_coord_lon (`Union[str, float]`, **optional**): Longitude of station/stop coordinate of the trip's origin. The coordinate can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * origin_coord_name (`str`, **optional**): Name of the trip's origin if coordinate cannot be resolved to an address or poi. Defaults to `None`.
        * destination_id (`str`, **optional**): Specifies the station/stop ID of the destination for the trip. Such ID can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * destination_id_ext (`str`, **optional**): Deprecated. Please use destId as it supports external IDs. Specifies the external station/stop ID of the destination for the trip. Such ID can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * destination_coord_lat (`Union[str, float]`, **optional**): Latitude of station/stop coordinate of the trip's destination. The coordinate can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * destination_coord_lon (`Union[str, float]`, **optional**): Longitude of station/stop coordinate of the trip's destination. The coordinate can be retrieved from stop_by_name() or stop_by_coords(). Defaults to `None`.
        * destination_coord_name (`str`, **optional**): Name of the trip's destination if coordinate cannot be resolved to an address or poi. Defaults to `None`.
        * via (`str`, **optional**): Complex structure to provide multiple via points separated by semicolon. This structure is build like this: `viaId|waittime|viastatus|products|direct|sleepingCar|couchetteCoach|attributes`. Read more about this in HAFAS ReST Documentation. Defaults to `None`.
        * via_id (`str`, **optional**): ID of a station/stop used as a via for the trip. Specifying a via station forces the trip search to look for trips which must pass through this station. Such ID can be retrieved from stop_by_name() or stop_by_coords(). If `via` is used, `via_id` and `via_wait_time ` are having no effect. Defaults to `None`.
        * via_gis (`str`, **optional**): Complex structure to provide multiple GIS via locations separated by semicolon. This structure is build like this: `locationId|locationMode|transportMode|placeType|usageType|mode|durationOfStay`. Read more about this in HAFAS ReST Documentation. Defaults to `None`.
        * via_wait_time (`int`, **optional**): Defines the waiting time spent at via station in minutes. If `via` is used, `via_id` and `via_wait_time` are having no effect. Defaults to 0.
        * avoid (`str`, **optional**): Complex structure to provide multiple points to be avoided separated by semicolon. This structure is build like this: `avoidId|avoidstatus` avoidId: id, extId or altId of the avoid, mandatory avoidstatus: one of NPAVM (do not run through if this is a meta station), NPAVO (do not run through), NCAVM (do not change if this is a meta station), NCAVO (do not change), optional but defaults to NCAVM Example: Just define three avoids by extId: `avoid="801234;801235;801236"`. Defaults to `None`.
        * avoid_id (`str`, **optional**): ID of a station/stop to be avoided as transfer stop for the trip. Such ID can be retrieved from stop_by_name() or stop_by_coords(). If `avoid` is used, `avoid_id` has no effect. Defaults to `None`.
        * change_time_percent (`int`, **optional**): Configures the walking speed when changing from one leg of the journey to the next one. It extends the time required for changes by a specified percentage. A value of 200 doubles the change time as initially calculated by the system. In the response, change time is presented in full minutes. If the calculation based on changeTime-Percent does not result in a full minute, it is rounded using "round half up" method. Defaults to `100`.
        * change_time_min (`int`, **optional**): Minimum change time at stop in minutes. Defaults to `None`.
        * change_time_max (`int`, **optional**): Maximum change time at stop in minutes. Defaults to `None`.
        * change_time_add (`int`, **optional**): This amount of minutes is added to the change time at each stop. Defaults to `None`.
        * change_max (`int`, **optional**): Maximum number of changes. In range 0-11. Defaults to `None`.
        * date (`Union[str, datetime]`, **optional**): Sets the start date for which the departures shall be retrieved. Represented in the format `YYYY-MM-DD` or as a datetime object. By default the current server date is used. Defaults to `None`.
        * time (`Union[str, datetime]`, **optional**): Sets the start time for which the departures shall be retrieved. Represented in the format `hh:mm[:ss]` in 24h nomenclature or as a datetime object. Seconds will be ignored for requests. By default the current server time is used. Defaults to `None`.
        * search_arrival (`bool`, **optional**): If set, the date and time parameters specify the arrival time for the trip search instead of the departure time. Defaults to `False`.
        * trips_after_time (`int`, **optional**): Minimum number of trips after the search time. Sum of `trips_after_time` and `trips_before_time` has to be less or equal 6. Read more about this in HAFAS ReST Documentation. In range 1-6. Defaults to `None`.
        * trips_before_time (`int`, **optional**): Minimum number of trips before the search time. Sum of `trips_after_time` and `trips_before_time` has to be less or equal 6. Read more about this in HAFAS ReST Documentation. In range 0-6. Defaults to `None`.
        * context (`str`, **optional**): Defines the starting point for the scroll back or forth operation. Use the scrB value from a previous result to scroll backwards in time and use the scrF value to scroll forth. Defaults to `None`.
        * passlist (`bool`, **optional**): Enables/disables the return of the passlist for each leg of the trip. Defaults to `False`.
        * operators (`Union[str, list]`, **optional**): Only trips provided by the given operators are part of the result. If the operator should not be part of the be trip, negate it by putting ! in front of it. Example: Filter for operator A and B: `operators=["A","B"]`. Defaults to `None`.
        * lines (`Union[str, list]`, **optional**): Only journeys running the given line are part of the result. If the line should not be part of the be trip, negate it by putting ! in front of it. Defaults to `None`.
        * lineids (`Union[str, list]`, **optional**): Only journeys running the given line (identified by its line ID) are part of the result. If the line should not be part of the be trip, negate it by putting ! in front of it. Defaults to `None`.
        * iv_include (`bool`, **optional**): Enables/disables search for individual transport routes. Defaults to `False`.
        * iv_only (`bool`, **optional**): Enables/disables search for individual transport routes only. Defaults to `False`.
        * bike_carriage (`bool`, **optional**): Enables/disables search for trips explicit allowing bike carriage. This will only work in combination with `change_max=0` as those trips are always meant to be direct connections. Defaults to `False`.
        * passing_points (`bool`, **optional**): Enables/disables the return of stops having no alighting and boarding in its passlist for each leg of the trip. Needs passlist enabled. Defaults to `False`.
        * real_time_mode (`Literal[RealTimeMode.FULL, RealTimeMode.INFOS, RealTimeMode.OFF, RealTimeMode.REALTIME, RealTimeMode.SERVER_DEFAULT]`, **optional**): Set the realtime mode to be used. Read more about this in HAFAS ReST Documentation. Defaults to `None`.
        * include_earlier (`bool`, **optional**): Disables search optimization in relation of duration. Defaults to `False`.
        * ict_alternatives (`bool`, **optional**): Enables/disables the search for alternatives with individualized change times (ICT). Defaults to `False`.
        * tariff (`bool`, **optional**): Enables/disables the output of tariff data. The default is configurable via provisioning. Defaults to `None`.
        * messages (`bool`, **optional**): Enables/disables the output of traffic messages. The default is configurable via provisioning. Defaults to `False`.
        * frequency (`bool`, **optional**): Enables/disables the calculation of frequency information. Defaults to `True`.

    ### Returns:
        * List[Trip]: List of Trip objects. Empty list if none found.
    """    

    if real_time_mode == None:
        real_time_mode = None
    else:
        real_time_mode = real_time_mode.code

    trips = []
    trips_raw = raw_trip_find(

        accessId=access_id,
        lang=lang.code,

        originId=origin_id,
        originExtId=origin_id_ext,
        originCoordLat=origin_coord_lat,
        originCoordLong=origin_coord_lon,
        originCoordName=origin_coord_name,

        destId=destination_id,
        destExtId=destination_id_ext,
        destCoordLat=destination_coord_lat,
        destCoordLong=destination_coord_lon,
        destCoordName=destination_coord_name,

        via=via,
        viaId=via_id,
        viaGis=via_gis,
        viaWaitTime=via_wait_time,

        avoid=avoid,
        avoidId=avoid_id,

        changeTimePercent=change_time_percent,
        minChangeTime=change_time_min,
        maxChangeTime=change_time_max,
        addChangeTime=change_time_add,
        maxChange=change_max,

        date=date,
        time=time,
        
        searchForArrival=search_arrival,

        numF=trips_after_time,
        numB=trips_before_time,

        context=context,

        passlist=passlist,
        operators=operators,
        
        lines=lines,
        lineids=lineids,

        includeIv=iv_include,
        ivOnly=iv_only,
        
        bikeCarriage=bike_carriage,

        showPassingPoints=passing_points,

        rtMode=real_time_mode,

        includeEarlier=include_earlier,
        withICTAlternatives=ict_alternatives,
        tariff=tariff,
        trafficMessages=messages,
        withFreq=frequency
    )

    find_exception(trips_raw)

    for trip in trips_raw["Trip"]:
        trips.append(Trip(trip))

    return trips


from enum import auto
from .auto_name import AutoName

class RealTimeMode(AutoName):
    """Enumeration used to declare types of real-time traffic."""

    FULL = auto()
    "Combined search on planned and real-time data This search consists of two steps: i. Search on scheduled data ii. If the result of step (i) contains a non-feasible connection, a search on real-time data is performed and all results are combined"

    INFOS = auto()
    "Search on planned data, use real-time information for display only: Connections are computed on the basis of planned data. Delays and feasibility of the connections are integrated into the result. Note that additional trains (supplied via realtime feed) will not be part of the resulting connections."

    OFF = auto()
    "Search on planned data, ignore real-time information completely: Connections are computed on the basis of planned data. No real-time information is shown."

    REALTIME = auto()
    "Search on real-time data: Connections are computed on the basis of real-time data, using planned schedule only whenever no real-time data is available. All connections computed are feasible with respect to the currently known real-time situation. Additional trains (supplied via real-time feed) will be found if these are part of a fast, comfortable, or direct connection (or economic connection, if economic search is activated)."

    SERVER_DEFAULT = auto()
    "One of the other real-times modes used by default for RMV."


from enum import auto
from .auto_name import AutoName

class FilterMode(AutoName):
    """Enumeration used to declare filters for nearby searches.  
    
    * DIST_PERI - Accentuate matches. Matches in the radius are first
    * EXCL_PERI - Returns matches inside the radius only
    * SLCT_PERI - Matches in the radius are excluded. Returns matches outside the radius only
    """

    DIST_PERI = auto()
    "Accentuate matches. Matches in the radius are first."

    EXCL_PERI = auto()
    "Returns matches inside the radius only."

    SLCT_PERI = auto()
    "Matches in the radius are excluded. Returns matches outside the radius only."


from enum import auto
from .auto_name import AutoNameProduct

class Product(AutoNameProduct):
    """Enumeration used to declare types of transport."""

    ICE = auto()
    "The Intercity Express (commonly known as ICE) is a system of high-speed trains predominantly running in Germany."

    IC = auto()
    "InterCity (commonly abbreviated IC on timetables and tickets) is the classification applied to certain long-distance passenger train services in Europe. Such trains (in contrast to regional, local, or commuter trains) generally call at major stations only."

    EC = auto()
    "EuroCity, abbreviated as EC, is a cross-border train category within the European inter-city rail network. In contrast to trains allocated to the lower-level \"IC\" (InterCity) category, EC trains are international services that meet 20 criteria covering comfort, speed, food service, and cleanliness."

    R = auto()
    "Regional rail, also known as local trains and stopping trains, are passenger rail services that operate between towns and cities. These trains operate with more stops over shorter distances than inter-city rail, but fewer stops and faster service than commuter rail."

    RB = auto()
    "The Regionalbahn (lit. Regional train; abbreviated RB) is a type of local passenger train (stopping train) in Germany. It is similar to the Regionalzug (R) and Regio (R) train categories in neighboring Austria and Switzerland, respectively."

    RE = auto()
    "In Germany, Luxembourg and Austria, the Regional-Express (RE, or in Austria: REX) is a type of regional train. It is similar to a semi-fast train, with average speed at about 70-90 km/h (top speed often 160 km/h) as it calls at fewer stations than Regionalbahn or S-Bahn trains, but stops more often than InterCity services."

    SBAHN = auto()
    "The S-Bahn is the name of hybrid urban-suburban rail systems serving a metropolitan region in German-speaking countries. Some of the larger S-Bahn systems provide service similar to rapid transit systems, while smaller ones often resemble commuter or even regional rail. The term derives from Schnellbahn, Stadtbahn or Stadtschnellbahn."

    UBAHN = auto()
    "A U-Bahn (short for subway, underground express train or metropolitan) is a usually underground, rail-bound means of transport for urban public transport (ÖPNV, city transport)."

    TRAM = auto()
    "A tram (also known as a streetcar or trolley in North America) is a rail vehicle that travels on tramway tracks on public urban streets; some include segments on segregated right-of-way."

    BUS = auto()
    "A bus (contracted from omnibus, with variants multibus, motorbus, autobus, etc.) is a road vehicle that carries significantly more passengers than an average car or van."

    BUS2 = auto()
    "Also a bus, but now idea why it exists anyway."

    FERRY = auto()
    "A ferry is a ship used to carry passengers, and sometimes vehicles and cargo, across a body of water."

    TAXI = auto()
    "A taxi, also known as a taxicab or simply a cab, is a type of vehicle for hire with a driver, used by a single passenger or small group of passengers, often for a non-shared ride."

    BAHN = auto()
    "Rail transport (also known as train transport) is a means of transport that transfers passengers and goods on wheeled vehicles running on rails, which are located on tracks. In contrast to road transport, where the vehicles run on a prepared flat surface, rail vehicles (rolling stock) are directionally guided by the tracks on which they run."


from pyrmv.classes.Stop import Stop
from datetime import datetime
from isodate import parse_duration

class Url():
    """Traffic message channel url object."""    

    def __init__(self, data: dict) -> None:
        self.name = data["name"]
        self.url = data["url"]

    def __str__(self) -> str:
        return f"{self.name}: {self.url}"

class Channel():
    """Traffic message channel object."""    

    def __init__(self, data: dict) -> None:
        self.name = data["name"]
        url = []
        for link in url:
            url.append(Url(link))
        self.url = url
        self.time_start = datetime.strptime(data["validFromTime"], "%H:%M:%S")
        self.date_start = datetime.strptime(data["validFromDate"], "%Y-%m-%d")
        self.time_end = datetime.strptime(data["validToTime"], "%H:%M:%S")
        self.date_end = datetime.strptime(data["validToDate"], "%Y-%m-%d")

    def __str__(self) -> str:
        return f"{self.name}: from {self.time_start} {self.date_start} until {self.time_end} {self.date_end}"


class Message():
    """Traffic message object."""    

    def __init__(self, data: dict) -> None:

        self.affected_stops = []
        if "affectedStops" in data:
            for stop in data["affectedStops"]["StopLocation"]:
                self.affected_stops.append(Stop(stop))

        if "validFromStop" in data:
            self.valid_from_stop = Stop(data["validFromStop"])
        else:
            self.valid_from_stop = None
        
        if "validToStop" in data:
            self.valid_to_stop = Stop(data["validToStop"])
        else:
            self.valid_to_stop = None

        self.channels = []
        for channel in data["channel"]:
            self.channels.append(Channel(channel))

        self.id = data["id"]
        self.active = data["act"]
        self.head = data["head"]
        self.lead = data["lead"]
        self.text = data["text"]
        self.company = data["company"]
        self.category = data["category"]
        self.priority = data["priority"]
        self.products = data["products"]
        self.icon = data["icon"]
        self.time_start = datetime.strptime(data["sTime"], "%H:%M:%S")
        self.date_start = datetime.strptime(data["sDate"], "%Y-%m-%d")
        self.time_end = datetime.strptime(data["eTime"], "%H:%M:%S")
        self.date_end = datetime.strptime(data["eDate"], "%Y-%m-%d")
        self.date_start_alt = data["altStart"]
        self.date_end_alt = data["altEnd"]
        self.time_modified = datetime.strptime(data["modTime"], "%H:%M:%S")
        self.date_modified = datetime.strptime(data["modDate"], "%Y-%m-%d")
        self.daily_start = datetime.strptime(data["dailyStartingAt"], "%H:%M:%S")
        self.daily_duration = parse_duration(data["dailyDuration"])

        if "baseType" in data:
            self.base_type = data["baseType"]
        else:
            self.base_type = None

    def __str__(self) -> str:
        return f"{self.base_type} message with priority {self.products} valid from {self.time_start.time()} {self.date_start.date()} until {self.time_end.time()} {self.date_end.date()}: {self.head} - {self.lead}"


class SotAlreadyArrivedError(Exception):
    """
    Trip already arrived.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SotNotStartedError(Exception):
    """
    Trip not started.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SotCancelledError(Exception):
    """
    Trip cancelled.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SotAllTrainsFilteredError(Exception):
    """
    All trips filtered.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__

class SotStayOnTripError(Exception):
    """
    No change. Stay on trip.
    """
    def __init__(self):
        super().__init__(self.__doc__)

    def __str__(self):
        return self.__doc__